"""
Test object detection operator
"""
import os
import sys
sys.path.append(os.getcwd())

from itertools import combinations

import numpy as np
import tensorflow as tf

from tests.utils import almost_equal, generate_object_detection_model, generate_data

from xplique.attributions import Occlusion, SmoothGrad
from xplique.commons.operators import object_detection_operator


def test_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    # 3 bounding boxes, one for each input sample
    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    explainer = Occlusion(model, operator=object_detection_operator, patch_size=4, patch_stride=2)

    # test with only one box to explain by image (3, 7)
    phis = explainer(x, obj_ref)
    assert phis.shape == (obj_ref.shape[0], input_shape[0], input_shape[1])
    assert phis[0, 0, 0] != np.nan

    phis2 = explainer(x, tf.expand_dims(obj_ref, axis=1))
    assert phis.shape == phis2.shape
    assert almost_equal(phis, phis2)


def test_gradient_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    explainer = SmoothGrad(model, nb_samples=10, operator=object_detection_operator)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    phis = explainer(x, obj_ref)

    assert phis.shape[:3] == (obj_ref.shape[0], input_shape[0], input_shape[1])


def test_several_boxes_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    explainer = SmoothGrad(model, nb_samples=10, operator=object_detection_operator)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    phis = explainer(x, obj_ref)
    assert phis.shape[:3] == (obj_ref.shape[0], input_shape[0], input_shape[1])

    several_object_refs = tf.tile(tf.expand_dims(obj_ref, axis=1), [1, 5, 1])
    
    phis = explainer(x, several_object_refs)

    assert phis.shape[:3] == (several_object_refs.shape[0], input_shape[0], input_shape[1])


def test_all_object_detector_operators():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    # set params
    parameters_normal = {
        "include_detection_probability": True,
        "include_classification_score": True,
    }

    parameters_intersection = {
        "include_detection_probability": False,
        "include_classification_score": False,
    }

    parameters_probability = {
        "include_detection_probability": True,
        "include_classification_score": False,
    }

    parameters_classification = {
        "include_detection_probability": False,
        "include_classification_score": True,
    }

    # create operators
    normal_op = lambda model, inputs, targets: \
            object_detection_operator(model, inputs, targets, **parameters_normal)

    intersection_op = lambda model, inputs, targets: \
            object_detection_operator(model, inputs, targets, **parameters_intersection)

    probability_op = lambda model, inputs, targets: \
            object_detection_operator(model, inputs, targets, **parameters_probability)

    classification_op = lambda model, inputs, targets: \
            object_detection_operator(model, inputs, targets, **parameters_classification)

    # compute explanations
    phis = Occlusion(model, operator=object_detection_operator, patch_size=4, patch_stride=2)(x, obj_ref)

    normal_phis = Occlusion(model, operator=normal_op, patch_size=4, patch_stride=2)(x, obj_ref)

    intersection_phis = Occlusion(model, operator=intersection_op, patch_size=4, patch_stride=2)(x, obj_ref)

    probability_phis = Occlusion(model, operator=probability_op, patch_size=4, patch_stride=2)(x, obj_ref)

    classification_phis = Occlusion(model, operator=classification_op, patch_size=4, patch_stride=2)(x, obj_ref)

    for phi in [phis, normal_phis, intersection_phis, probability_phis, classification_phis]:
        assert phi.shape[:3] == (obj_ref.shape[0], input_shape[0], input_shape[1])
    
    assert almost_equal(phis, normal_phis)

    for phi1, phi2 in combinations([normal_phis, intersection_phis, probability_phis, classification_phis], 2):
        assert not almost_equal(phi1, phi2)


"""
Representativity & Consistency metric
"""

import numpy as np
import tensorflow as tf
from scipy.stats import spearmanr

from ..commons import numpy_sanitize
from ..types import Callable, Optional, Union, Tuple, Dict


class MeGe:
    """
    Used to calculate metrics for representativity and consistency of explanations.
    MeGe (representativity) gives you an overview of the generalization of your explanations:
    how much seeing one explanation informs you about the others.
    ReCo (consistency) is the consistency score of the explanations, it informs about the
    confidence of the explanations, how much two models explanations will not contradict
    each other.

    Ref. Fel, Vigouroux & al., How good is your explanation? Algorithmic stability measures to
    assess the quality of explanations for deep neural networks (2020).
    https://arxiv.org/abs/2009.04521

    Parameters
    ----------
    learning_algorithm
        Function that will be called with (x_train, y_train, x_test, y_test) and that should
        return a Keras Model.
    inputs
        Input samples under study.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to explain at once, if None compute all at once.
    k_splits
        Number of splits to estimate the metrics (usually between 4 and 6 are enough).
    """
    # pylint: disable=R0902

    def __init__(self,
                 learning_algorithm: Callable,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 k_splits: int = 4):
        self.inputs, self.targets = numpy_sanitize(inputs, targets)
        self.batch_size = batch_size
        self.k_splits = k_splits

        # start by splitting the dataset in K splits, and train K model on K-1 split
        assert len(inputs) % k_splits == 0, "`k_splits` must divide the length of the dataset."

        self.split_len = len(inputs) // k_splits
        self.x_splits = np.array([
            self.inputs[i * self.split_len:(i + 1) * self.split_len] for i in range(k_splits)])
        self.y_splits = np.array([
            self.targets[i * self.split_len:(i + 1) * self.split_len] for i in range(k_splits)])

        self.models = []
        for i in range(k_splits):
            # model_i is trained with the split k \ k_i (without k_i)
            ids_to_keep = [j for j in range(k_splits) if j != i]
            x_train = self.x_splits[ids_to_keep].reshape((-1, *self.inputs.shape[1:]))
            y_train = self.y_splits[ids_to_keep].reshape((-1, *self.targets.shape[1:]))

            model = learning_algorithm(x_train, y_train, self.x_splits[i], self.y_splits[i])
            self.models.append(model)

    def evaluate(self,
                 explainer_class: Callable,
                 explainer_params: Optional[Dict] = None) -> Tuple[float, float]:
        # pylint: disable=C0103,R1702
        """
        Evaluate the MeGe score.

        Parameters
        ----------
        explainer_class
            Explainer Attribution class to use.
        explainer_params
            Explainer dict parameters to use to create each explainer.

        Returns
        -------
        mege
            The Mean representativity score.
        reco
            The Relative Consistency score.
        """
        explainer_params = explainer_params or {}

        # for each model, we will need his explanation and his predictions
        predictions = np.array([
            np.argmax(model.predict(self.inputs, batch_size=self.batch_size), -1)
            for model in self.models
        ])
        explanations = np.array([
            explainer_class(model, **explainer_params)(self.inputs, self.targets)
            for model in self.models
        ])

        s_eq, s_ne = self._pairwise_distances(predictions, explanations)
        s_total = np.concatenate([s_eq, s_ne], axis=0)

        mege = 1.0 / (1.0 + np.mean(s_eq))

        reco = 0.0
        step = (len(s_total) // 20) + 1 # skip some element to speed up computation
        for gamma in np.sort(s_total)[::step]:
            tp = np.sum(s_eq < gamma)
            tn = np.sum(s_ne > gamma)
            fp = np.sum(s_ne < gamma)
            fn = np.sum(s_eq > gamma)

            tpr = tp / (tp + fn + 1e-10)
            tnr = tn / (tn + fp + 1e-10)

            reco_gamma = tpr + tnr - 1.0
            reco = np.max([reco, reco_gamma])

        return mege, reco

    def _pairwise_distances(self,
                            predictions: np.array,
                            explanations: np.array) -> Tuple[np.array, np.array]:
        """
        Compute all the pairwise distance between the models explanations.
        Use those distances to build two sets: s_equal for the distance between explanations of
        the same predictions, and s_not_equal for the distances between explanations of
        different predictions.

        Parameters
        ----------
        predictions
            Models predictions (one for each explanation).
        explanations
            Models explanations (one for each prediction).

        Returns
        -------
        s_eq
            The set of distances when the predictions are the same.
        s_ne
            The set of distances when the predictions are different.
        """
        # pylint: disable=C0103,R1702
        # compute pairwise distances and add them into correct distribution
        # s_eq is S^= (eq.3) and s_ne is S^{!=} (eq.4)
        s_eq, s_ne = [], []

        for i in range(self.k_splits):
            for j in range(i+1, self.k_splits):
                # get all the explanations for the pairs (model_i, model_j)
                # with i != j
                e_i = explanations[i]
                e_j = explanations[j]

                for n, e_i_n in enumerate(e_i):
                    # compute only the distance between one model
                    # that has seen the data and the other who don't
                    # to ensure that, the split_id must be i or j
                    # that would mean that one the model has use this split to test
                    split_id = n // self.split_len
                    compute_distance = split_id in [i, j]

                    if compute_distance:
                        pred_i = predictions[i, n]
                        pred_j = predictions[j, n]

                        dist = self._spearman_distance(e_i_n, e_j[n])

                        # we need at least one good answer
                        if pred_i == np.argmax(self.targets[n], -1) or \
                           pred_j == np.argmax(self.targets[n], -1):
                            # both are correct, add to S=
                            if pred_i == pred_j:
                                s_eq.append(dist)
                            # one is correct, add to S!=
                            else:
                                s_ne.append(dist)

        # ignore possible nan values (e.g grad-cam is full of zero)
        s_eq = np.array(s_eq)[~np.isnan(s_eq)]
        s_ne = np.array(s_ne)[~np.isnan(s_ne)]

        return s_eq, s_ne

    @staticmethod
    def _spearman_distance(explanation_a: Union[tf.Tensor, np.array],
                           explanation_b: Union[tf.Tensor, np.array]) -> float:
        """
        Compute the spearman distance between two explanations.

        Parameters
        ----------
        explanation_a
            First explanation as tensor or numpy array.
        explanation_b
            Second explanation as tensor or numpy array.

        Returns
        -------
        dist
            The spearman distance defined as (1 - |spearman_corr(., .)|)**0.5
        """
        explanation_a = MeGe._sanitize_explanation(explanation_a)
        explanation_b = MeGe._sanitize_explanation(explanation_b)

        rho, _ = spearmanr(explanation_a, explanation_b)
        dist = np.sqrt(1.0 - np.abs(rho))
        return dist


    @staticmethod
    def _sanitize_explanation(explanation: Union[tf.Tensor, np.array]) -> np.array:
        """
        Sanitize explanations to ensure we can properly compute the spearman distance.

        Parameters
        ----------
        explanation
            Explanation to sanitize.

        Returns
        -------
        explanation
            Explanation sanitized.
        """
        if len(explanation.shape) > 2:
            explanation = np.mean(explanation, -1)
        explanation = np.nan_to_num(explanation, 0.0).flatten()

        return explanation



"""
Module related to Object detector method
"""

from deprecated import deprecated

import numpy as np
import tensorflow as tf

from ..types import Optional, Callable, Union

from .base import BlackBoxExplainer, WhiteBoxExplainer
from ..commons import get_gradient_functions
from ..commons.operators import object_detection_operator
from ..utils_functions.object_detection import _box_iou, _format_objects


OLD_OBJECT_DETECTION_DEPRECATION_MESSAGE = """
\n
The method to compute attribution explanation explanations changed drastically after version 1.0.0. 
For more information please refer to the documentation:
https://deel-ai.github.io/xplique/latest/api/attributions/object_detection/

Nonetheless, here is a quick example of how it is used now:
```
from xplique.attributions import AnyMethod
explainer = AnyMethod(model, operator="object detection", ...)
explanation = explainer(inputs, targets)  # be aware, targets are specific to object detection here
```
"""


class BoundingBoxesExplainer(BlackBoxExplainer):
    """
    For a given black box explainer, this class allows to find explications of an object detector
    model. The object model detector shall return a list (length of the size of the batch)
    containing a tensor of 2 dimensions.
    The first dimension of the tensor is the number of bounding boxes found in the image
    The second dimension is:
    [xmin, ymin, xmax, ymax, probability_detection, ones_hot_classif_result]

    This work is a generalization of the following article at any kind of black box explainer and
    also can be used for other kind of object detector (like segmentation)

    Ref. Petsiuk & al., Black-box Explanation of Object Detectors via Saliency Maps (2021).
    https://arxiv.org/pdf/2006.03204.pdf

    Parameters
    ----------
    explainer
        the black box explainer used to explain the object detector model
    _
        inheritance from old versions
    intersection_score
        the iou calculator used to compare two objects.
    """

    @deprecated(version="1.0.0", reason=OLD_OBJECT_DETECTION_DEPRECATION_MESSAGE)
    def __init__(self,
                 explainer: BlackBoxExplainer,
                 _: Optional[Callable] = _format_objects,
                 intersection_score: Optional[Callable]  = _box_iou):
        # make operator function based on arguments
        operator = lambda model, inputs, targets: \
            object_detection_operator(model, inputs, targets, intersection_score)

        # BlackBoxExplainer init to set operator
        super().__init__(model=explainer.model, batch_size=explainer.batch_size, operator=operator)
        self.explainer = explainer

        # update explainer inference functions for explain method
        self.explainer.inference_function = self.inference_function
        self.explainer.batch_inference_function = self.batch_inference_function

        if isinstance(self.explainer, WhiteBoxExplainer):
            # check and get gradient function from model and operator
            self.gradient, self.batch_gradient = get_gradient_functions(self.model, operator)
            self.explainer.gradient = self.gradient
            self.explainer.batch_gradient = self.batch_gradient

    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.array],
                targets: Optional[Union[tf.Tensor, np.array]] = None) -> tf.Tensor:
        """
        Compute the explanation of the object detection through the explainer

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, ...). With features matching the object formatting.
            See object detection operator documentation for more information
            More information in the documentation.

        Returns
        -------
        explanation
            The resulting object detection explanation
        """
        if len(tf.shape(targets)) == 1:
            targets = tf.expand_dims(targets, axis=0)

        return self.explainer.explain(inputs, targets)



import tensorflow as tf

from xplique.utils_functions.segmentation import *

def get_prediction():
    predictions = [[[0.6, 0.6, 0.6, 0.2, 0.2],
                    [0.6, 0.6, 0.2, 0.2, 0.2],
                    [0.6, 0.2, 0.2, 0.2, 0.2],
                    [0.2, 0.2, 0.2, 0.2, 0.2],
                    [0.2, 0.2, 0.2, 0.2, 0.2],],
                   [[0.2, 0.2, 0.2, 0.6, 0.6],
                    [0.2, 0.2, 0.6, 0.6, 0.6],
                    [0.2, 0.6, 0.6, 0.6, 0.2],
                    [0.6, 0.6, 0.6, 0.2, 0.2],
                    [0.6, 0.6, 0.2, 0.2, 0.2],],
                   [[0.2, 0.2, 0.2, 0.2, 0.2],
                    [0.2, 0.2, 0.2, 0.2, 0.2],
                    [0.2, 0.2, 0.2, 0.2, 0.6],
                    [0.2, 0.2, 0.2, 0.6, 0.6],
                    [0.2, 0.2, 0.6, 0.6, 0.6],],]
    
    predictions = tf.convert_to_tensor(predictions, tf.float32)

    predictions = tf.transpose(predictions, perm=[1, 2, 0])

    assert tf.reduce_all(tf.equal(tf.reduce_sum(predictions, axis=-1), tf.ones((5, 5))))

    return predictions


def test_get_class_zone():
    predictions = get_prediction()

    target_1 = get_class_zone(predictions, class_id=1)

    expected_target = tf.transpose(tf.convert_to_tensor(
        [[[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],
         [[0, 0, 0, 1, 1],
          [0, 0, 1, 1, 1],
          [0, 1, 1, 1, 0],
          [1, 1, 1, 0, 0],
          [1, 1, 0, 0, 0],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],],tf.float32), perm=[1, 2, 0])

    assert tf.reduce_all(tf.equal(target_1, expected_target))
    
    target_0 = get_class_zone(predictions, class_id=0)
    target_2 = get_class_zone(predictions, class_id=2)

    assert tf.reduce_all(tf.equal(
        tf.reduce_sum(tf.stack([target_0, target_1, target_2], axis=0), axis=[0, 3]),
        tf.fill((5, 5), 1.0)))


def test_get_connected_zone():
    predictions = get_prediction()

    target_1 = get_connected_zone(predictions, coordinates=(2, 2))

    expected_target = tf.transpose(tf.convert_to_tensor(
        [[[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],
         [[0, 0, 0, 1, 1],
          [0, 0, 1, 1, 1],
          [0, 1, 1, 1, 0],
          [1, 1, 1, 0, 0],
          [1, 1, 0, 0, 0],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],],tf.float32), perm=[1, 2, 0])

    assert tf.reduce_all(tf.equal(target_1, expected_target))
    
    target_0 = get_connected_zone(predictions, coordinates=(0, 0))
    target_2 = get_connected_zone(predictions, coordinates=(4, 4))

    assert tf.reduce_all(tf.equal(
        tf.reduce_sum(tf.stack([target_0, target_1, target_2], axis=0), axis=[0, 3]),
        tf.fill((5, 5), 1.0)))


def test_list_class_connected_zones():
    predictions = get_prediction()

    predictions = tf.stack([predictions[:, :, 0] + predictions[:, :, 2], predictions[:, :, 1]], axis=-1)

    zones_0 = list_class_connected_zones(predictions, class_id=0, zone_minimum_size=1)
    zones_1 = list_class_connected_zones(predictions, class_id=1, zone_minimum_size=1)
    no_zones = list_class_connected_zones(predictions, class_id=0, zone_minimum_size=10)

    assert len(zones_0) == 2
    assert len(zones_1) == 1
    assert len(no_zones) == 0

    expected_zones_1 = tf.transpose(tf.convert_to_tensor(
        [[[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],
         [[0, 0, 0, 1, 1],
          [0, 0, 1, 1, 1],
          [0, 1, 1, 1, 0],
          [1, 1, 1, 0, 0],
          [1, 1, 0, 0, 0],],], tf.float32), perm=[1, 2, 0])
    
    assert tf.reduce_all(tf.equal(zones_1[0], expected_zones_1))

    expected_zones_21 = tf.transpose(tf.convert_to_tensor(
        [[[1, 1, 1, 0, 0],
          [1, 1, 0, 0, 0],
          [1, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],], tf.float32), perm=[1, 2, 0])

    expected_zones_22 = tf.transpose(tf.convert_to_tensor(
        [[[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 1],
          [0, 0, 0, 1, 1],
          [0, 0, 1, 1, 1],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],], tf.float32), perm=[1, 2, 0])
    
    assert tf.reduce_all(tf.equal(zones_0[0], expected_zones_21))\
        or tf.reduce_all(tf.equal(zones_0[0], expected_zones_22))
    
    assert tf.reduce_all(tf.equal(zones_0[1], expected_zones_21))\
        or tf.reduce_all(tf.equal(zones_0[1], expected_zones_22))



def test_get_in_out_border():
    predictions = get_prediction()

    central_zone = get_connected_zone(predictions, coordinates=(2, 2))

    borders = get_in_out_border(central_zone)

    expected_borders = tf.transpose(tf.convert_to_tensor(
        [[[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],
         [[ 0, -1, -1,1,  0],
          [-1, -1, 1, 1, 1],
          [-1, 1, 1,1 , -1],
          [ 1, 1, 1, -1, -1],
          [ 0,1, -1, -1,  0],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],],tf.float32), perm=[1, 2, 0])

    assert tf.reduce_all(tf.equal(borders, expected_borders))
    

def test_get_common_border():
    predictions = get_prediction()

    left_corner_zone = get_connected_zone(predictions, coordinates=(0, 0))
    central_zone = get_connected_zone(predictions, coordinates=(2, 2))

    left_corner_borders = get_in_out_border(left_corner_zone)
    central_borders = get_in_out_border(central_zone)

    common_borders_0 = get_common_border(left_corner_borders, central_borders)
    common_borders_1 = get_common_border(central_borders, left_corner_borders)

    expected_common_borders = tf.transpose(tf.convert_to_tensor(
        [[[ 0,  1,  1, -1,  0],
          [ 1,  1, -1, -1,  0],
          [ 1, -1, -1,  0,  0],
          [-1, -1,  0,  0,  0],
          [ 0,  0,  0,  0,  0],],
         [[ 0, -1, -1,  1,  0],
          [-1, -1,  1,  1,  0],
          [-1,  1,  1,  0,  0],
          [ 1,  1,  0,  0,  0],
          [ 0,  0,  0,  0,  0],],
         [[0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],
          [0, 0, 0, 0, 0],],],tf.float32), perm=[1, 2, 0])

    assert tf.reduce_all(tf.equal(common_borders_0, common_borders_1))

    assert tf.reduce_all(tf.equal(common_borders_0, expected_common_borders))


import tensorflow as tf
import numpy as np

from xplique.commons import tensor_sanitize, numpy_sanitize

from ..utils import generate_data


def test_tensor_sanitize():
    """Ensure we get tf.Tensor for numpy array, tf tensor and tf.data.Dataset"""
    nb_samples = 71
    inputs_shapes = [
        (32, 32, 1), (32, 32, 3)
    ]

    for shape in inputs_shapes:
        inputs_np, targets_np = generate_data(shape, 10, nb_samples)
        inputs_tf, targets_tf = tf.cast(inputs_np, tf.float32), tf.cast(targets_np, tf.float32)
        dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np))
        dataset_batched = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np)).batch(10)

        for inputs, targets in [
            (inputs_np, targets_np),
            (inputs_tf, targets_tf),
            (dataset, None),
            (dataset_batched, None),
        ]:
            inputs_sanitize_tf, targets_sanitize_tf = tensor_sanitize(inputs, targets)
            inputs_sanitize_np, targets_sanitize_np = numpy_sanitize(inputs, targets)

            assert isinstance(inputs_sanitize_tf, tf.Tensor)
            assert isinstance(targets_sanitize_tf, tf.Tensor)
            assert isinstance(inputs_sanitize_np, np.ndarray)
            assert isinstance(targets_sanitize_np, np.ndarray)

            assert len(inputs_sanitize_tf) == nb_samples
            assert len(targets_sanitize_tf) == nb_samples
            assert len(inputs_sanitize_np) == nb_samples
            assert len(targets_sanitize_np) == nb_samples



"""
Tests that BlackBox attribution model can indeed handle
several model type with a native User eXperience
"""

import numpy as np
import tensorflow as tf

from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier

from xplique.attributions import (Occlusion, Rise, Lime, KernelShap, SobolAttributionMethod)
from xplique.commons.operators_operations import predictions_operator, batch_predictions,\
    batch_predictions_one_hot_callable
from xplique.commons.callable_operations import predictions_one_hot_callable

from ..utils import generate_data, generate_model, generate_regression_model

def _default_methods_tabular(model):
    return [
        Occlusion(model, patch_size = 1, patch_stride = 1),
        Lime(model),
        KernelShap(model)
    ]

def _default_methods_images(model):
    return [
        Rise(model),
        Occlusion(model),
        Lime(model, map_to_interpret_space=_map_to_interpret_space),
        KernelShap(model, map_to_interpret_space=_map_to_interpret_space),
        SobolAttributionMethod(model, grid_size=2, nb_design=2)
    ]

def _map_to_interpret_space(inp):
    width = inp.shape[0]
    height = inp.shape[1]

    mapping = tf.range(width*height)
    mapping = tf.reshape(mapping, (width, height))
    mapping = tf.cast(mapping, tf.int32)

    return mapping

class DenseModule(tf.Module):
    """
    A generic tf.Module Dense layer
    """
    def __init__(self, input_size, nb_labels):
        super(DenseModule, self).__init__()
        self.weights = tf.Variable(
            tf.random.normal([input_size, nb_labels]), name='w')
        self.bias = tf.Variable(tf.zeros([nb_labels]), name='b')
    def __call__(self, x):
        output = tf.matmul(x, self.weights) + self.bias
        return tf.nn.relu(output)

class ImagesModule(tf.Module):
    """
    A generic tf.Module with flatten and two dense layers
    """
    def __init__(self, input_shape, nb_labels):
        super(ImagesModule, self).__init__()
        self.input_size = input_shape[0]*input_shape[1]*input_shape[2]
        self.dense1 = DenseModule(self.input_size, 64)
        self.dense2 = DenseModule(64, nb_labels)
    def __call__(self, x):
        x = tf.reshape(x, [x.shape[0],-1])
        x = self.dense1(x)
        output = self.dense2(x)
        return output

def _default_tf_model_tabular(input_size, nb_labels):
    tf_keras_model = generate_regression_model((input_size,), nb_labels)
    tf_keras_layer = tf.keras.layers.Dense(nb_labels)
    tf_keras_layer(tf.random.normal([32, input_size]))
    tf_module = DenseModule(input_size, nb_labels)

    return [
        tf_keras_model,
        tf_keras_layer,
        tf_module
    ]

def _default_tf_model_images(input_shape, nb_labels):

    tf_keras_model = generate_model(input_shape, nb_labels)

    # create a mode through tf functionnal api
    functional_inputs = tf.keras.Input(shape = input_shape)
    functional_flatten = tf.keras.layers.Flatten()
    functional_x = functional_flatten(functional_inputs)
    functional_x = tf.keras.layers.Dense(64, activation="relu")(functional_x)
    functional_outputs = tf.keras.layers.Dense(nb_labels)(functional_x)
    tf_functional_model = tf.keras.Model(inputs=functional_inputs, outputs=functional_outputs)

    tf_module = ImagesModule(input_shape, nb_labels)

    return [
        tf_keras_model,
        tf_functional_model,
        tf_module
    ]

def _default_callable_tabular(input_size, nb_labels):
    sk_svc = SVC(probability=True)
    sk_rf = RandomForestClassifier()
    sk_api = RandomSklearn(input_size, nb_labels)
    np_callable = RandomNpCallable(input_size, nb_labels)

    return [
        sk_svc,
        sk_rf,
        sk_api,
        np_callable
    ]

def _default_callable_images(input_shape, nb_labels):
    sk_api = RandomSklearn(input_shape, nb_labels)
    np_callable = RandomNpCallable(input_shape, nb_labels)

    return [
        sk_api,
        np_callable
    ]

class RandomSklearn():
    """
    A class with a fit and predict_proba attributes such as the one
    in sklearn api, which returns random probabilities
    """
    def __init__(self, input_shape, nb_labels):
        self.input_shape = input_shape
        self.nb_labels = nb_labels
    def fit(self, inputs, targets):
        pass
    def predict_proba(self, inputs):
        return np.random.random((inputs.shape[0], self.nb_labels))

class RandomNpCallable():
    """
    A class with a call defines on numpy array
    """
    def __init__(self, input_shape, nb_labels):
        self.input_shape = input_shape
        self.nb_labels = nb_labels
    def fit(self, inputs, targets):
        pass
    def __call__(self, inputs):
        return np.random.random((inputs.shape[0], self.nb_labels))

def test_tf_models_tabular():
    """
    Test if on tabular data, BlackBox methods work as expected with tf family of callable,
    at least the most commons.
    """
    input_size = 32
    nb_labels = 10
    samples = 10

    inputs, targets = generate_data((input_size,), nb_labels, samples)
    models = _default_tf_model_tabular(input_size, nb_labels)

    for model in models:
        explainers = _default_methods_tabular(model)

        for explainer in explainers:
            assert explainer.inference_function is predictions_operator
            assert explainer.batch_inference_function is batch_predictions

            explanations = explainer(inputs, targets)

            assert explanations.shape == (samples, input_size)

def test_tf_models_images():
    """
    Test if on images data, BlackBox methods work as expected with tf family of callable,
    at least the most commons.
    """
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10
    samples = 10

    for input_shape in input_shapes:
        inputs, targets = generate_data(input_shape, nb_labels, samples)
        models = _default_tf_model_images(input_shape, nb_labels)
        for model in models:

            explainers = _default_methods_images(model)

            for explainer in explainers:
                assert explainer.inference_function is predictions_operator
                assert explainer.batch_inference_function is batch_predictions

                explanations = explainer(inputs, targets)

                assert explanations.shape == (samples, *input_shape[:2])

def test_callable_models_tabular():
    """
    Test if on tabular data we can handle sklearn models, sklearn like models (with
    predict_proba attributes) and callable which requires a np.ndarray input
    """
    input_size = 32
    nb_labels = 10
    samples = 10

    inputs, targets = generate_data((input_size,), nb_labels, samples)
    models = _default_callable_tabular(input_size, nb_labels)

    for model in models:
        no_one_hot_targets = np.arange(0,10)

        model.fit(inputs, no_one_hot_targets)
        explainers = _default_methods_tabular(model)

        for explainer in explainers:
            assert explainer.inference_function is predictions_one_hot_callable
            assert explainer.batch_inference_function is batch_predictions_one_hot_callable

            explanations = explainer(inputs, targets)

            assert explanations.shape == (samples, input_size)

def test_callable_models_images():
    """
    Test if on images data we can handle sklearn like models (with predict_proba attributes)
    and callable which requires a np.ndarray input
    """
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10
    samples = 10

    for input_shape in input_shapes:
        inputs, targets = generate_data(input_shape, nb_labels, samples)
        models = _default_callable_images(input_shape, nb_labels)
        for model in models:

            explainers = _default_methods_images(model)

            for explainer in explainers:
                assert explainer.inference_function is predictions_one_hot_callable
                assert explainer.batch_inference_function is batch_predictions_one_hot_callable

                explanations = explainer(inputs, targets)

                assert explanations.shape == (samples, *input_shape[:2])



"""
Stochastic transformations
"""

import tensorflow as tf

from ..types import Tuple, Callable, List


def random_blur(sigma_range: Tuple[float, float] = (1.0, 2.0),
                kernel_size: int = 10) -> Callable:
    """
    Generate a function that apply a random gaussian blur to the batch.

    Parameters
    ----------
    sigma_range
        Min and max sigma (or scale) of the gaussian kernel.
    kernel_size
        Size of the gaussian kernel

    Returns
    -------
    blur
        Transformation function applying random blur.
    """
    uniform = tf.linspace(-(kernel_size - 1) / 2., (kernel_size - 1) / 2.,
                          kernel_size)
    uniform_xx, uniform_yy = tf.meshgrid(uniform, uniform)

    kernel_size = tf.cast(kernel_size, tf.float32)
    sigma_min = tf.cast(max(sigma_range[0], 0.1), tf.float32)
    sigma_max = tf.cast(max(sigma_range[1], 0.1), tf.float32)

    def blur(images: tf.Tensor) -> tf.Tensor:
        sigma = tf.random.uniform([], minval=sigma_min, maxval=sigma_max,
                                  dtype=tf.float32)

        kernel = tf.exp(-0.5 * (uniform_xx ** 2 + uniform_yy ** 2) / sigma ** 2)
        kernel /= tf.reduce_sum(kernel)

        kernel = tf.reshape(kernel, (kernel_size, kernel_size, 1, 1))
        kernel = tf.tile(kernel, [1, 1, 3, 1])

        return tf.nn.depthwise_conv2d(images, kernel, strides=[1, 1, 1, 1],
                                      padding='SAME')

    return blur


def random_jitter(delta: int = 6) -> Callable:
    """
    Generate a function that perform a random jitter on batch of images.

    Parameters
    ----------
    delta
        Max of the shift

    Returns
    -------
    jitter
        Transformation function applying random jitter.

    """

    def jitter(images: tf.Tensor) -> tf.Tensor:
        shape = tf.shape(images)
        images = tf.image.random_crop(images, (shape[0], shape[1] - delta, shape[2] - delta,
                                               shape[-1]))
        return images

    return jitter


def random_scale(scale_range: Tuple[float, float] = (0.95, 1.05)) -> Callable:
    """
    Generate a function that apply a random scaling to the batch. Preserve
    aspect ratio.

    Parameters
    ----------
    scale_range
        Min and max scaling factor.

    Returns
    -------
    scale
        Transformation function applying random scaling.
    """
    min_scale = tf.cast(scale_range[0], tf.float32)
    max_scale = tf.cast(scale_range[1], tf.float32)

    def scale(images: tf.Tensor) -> tf.Tensor:
        _, width, height, _ = images.shape
        scale_factor = tf.random.uniform([], minval=min_scale, maxval=max_scale,
                                         dtype=tf.float32)
        return tf.image.resize(images, tf.cast([width * scale_factor,
                                                height * scale_factor], tf.int32))

    return scale


def random_flip(horizontal: bool = True, vertical: bool = False) -> Callable:
    """
    Generate a function that apply random flip (horizontal/vertical) to the
    batch.

    Parameters
    ----------
    horizontal
        Whether to perform random horizontal flipping (left/right)
    vertical
        Whether to perform random vertical flipping (top/down)

    Returns
    -------
    flip
        Transformation function applying random flipping.
    """

    def flip(images: tf.Tensor) -> tf.Tensor:
        if horizontal:
            images = tf.image.random_flip_left_right(images)
        if vertical:
            images = tf.image.random_flip_up_down(images)
        return images

    return flip


def pad(size: int = 6, pad_value: float = 0.0) -> Callable:
    """
    Generate a function that apply padding to a batch of images.

    Parameters
    ----------
    size
        Size of the padding
    pad_value
        Value of the padded pixels

    Returns
    -------
    pad_func
        Transformation function applying padding.
    """
    pad_array = [(0, 0), (size, size), (size, size), (0, 0)]
    pad_value = tf.cast(pad_value, tf.float32)

    def pad_func(images: tf.Tensor) -> tf.Tensor:
        return tf.pad(images, pad_array, mode="CONSTANT", constant_values=pad_value)

    return pad_func


def compose_transformations(transformations: List[Callable]) -> Callable:
    """
    Return a function that combine all the transformations passed and resize
    the images at the end.

    Parameters
    ----------
    transformations
        List of transformations, like the one in this module.

    Returns
    -------
    composed_func
        The combinations of the functions passed and a resize.
    """

    def composed_func(images: tf.Tensor) -> tf.Tensor:
        for func in transformations:
            images = func(images)
        return images

    return composed_func


def generate_standard_transformations(size: int) -> Callable:
    """
    Prepare a set of (apparently) robust transformations.

    Parameters
    ----------
    size
        Input size of the image.

    Returns
    -------
    transformations
        A combinations of transformations to make the optimization robust.
    """
    unit = int(size / 16)

    return compose_transformations([
        pad(unit * 4, 0.0),
        random_jitter(unit * 2),
        random_jitter(unit * 2),
        random_jitter(unit * 4),
        random_jitter(unit * 4),
        random_jitter(unit * 4),
        random_scale((0.92, 0.96)),
        random_blur(sigma_range=(1.0, 1.1)),
        random_jitter(unit),
        random_jitter(unit),
        random_flip()
    ])



"""
Plots for metric
"""

import matplotlib
import matplotlib.pyplot as plt

from ..types import Optional, Union, Dict, Any


def barplot(
        scores: Dict[str, Dict[str, float]],
        sort_metric: str = None,
        ascending: bool = False,
        title: str = "",
        filepath: str = None,
        methods_colors: Optional[Union[Dict[str, Any], str]] = None
):
    """
    Make a bar chart gathering the score of each method for the different metrics used.

    Parameters
    ----------
    scores
        Dictionary of dictionary, with metrics scores,
        1st dict: the keys are the metric names and the values are dictionaries,
        2nd dict: the keys are the method names
        and the values the corresponding method-metric score.
    sort_metric
        Name of the metric used to sort the methods (a key of the directory).
    ascending
        Specify the sorting direction (need a sort_metric specified).
    title
        Title of the graphic.
    filepath
        Path the file will be saved at. If None, the function will call plt.show().
    methods_colors
        Either a dictionary of colors (key are method name, value are matplotlib
        supported color) or string of a matplotlib cmap (e.g 'Set3', 'jet'...).
    """
    # pylint: disable=invalid-name

    metrics = list(scores.keys())
    # sort values
    if sort_metric is not None:
        methods = [
            k for k, _ in
            sorted(scores[sort_metric].items(), key=lambda item: item[1], reverse=not ascending)
        ]
    else:
        # get methods from a metric (order do not import)
        methods = list(scores[list(scores.keys())[0]])

    # set x abscissa for each metric
    metrics_x = {metrics[i]: i for i in range(len(metrics))}

    # set x delta for each method
    width = (1 - 0.2) / len(methods)
    # -0.4 to get 0.2 between each metric, 0.5 width to center the bars
    methods_d = {methods[i]: -0.4 + (i + 0.5) * width for i in range(len(methods))}

    # set methods colors
    if not isinstance(methods_colors, dict):
        # either None or string
        if methods_colors is None:
            # default cmap
            cmap = matplotlib.cm.get_cmap("Set3")
        else:
            # methods_color is a string linking to a cmap
            cmap = matplotlib.cm.get_cmap(methods_colors)

        methods_colors = {methods[i]: cmap((i + 1) / len(methods))
                          for i in range(len(methods))}

    _, ax = plt.subplots(figsize=(4 + 2 * len(metrics), 4))
    handles = {}
    for metric in metrics:
        for method in methods:
            handles[method] = ax.bar(
                metrics_x[metric] + methods_d[method],
                scores[metric][method],
                color=methods_colors[method],
                align="center",
                width=width
            )

    ax.set_title(title)
    plt.legend(handles.values(), methods, title="attribution methods", bbox_to_anchor=(1.05, 1))
    ax.set_xlabel("metrics")
    ax.set_ylabel("score")
    plt.xticks(list(range(0, len(metrics))))
    ax.set_xticklabels(metrics)
    plt.xticks(rotation=0)
    plt.tight_layout()

    if filepath is not None:
        plt.savefig(filepath)
    else:
        plt.show()


def fidelity_curves(
        detailed_scores: Dict[str, Dict[int, float]],
        title: str = "",
        filepath: Optional[str] = None,
        methods_colors: Optional[Union[Dict[str, Any], str]] = None
):
    """
    Plot the evolution curves for Insertion and Deletion metrics
    based on their detailed_evaluate() method.
    Plot the evolution of a model score depending on the number features perturbed

    Parameters
    ----------
    detailed_scores
        Dictionary of methods detailed scores.
        The keys are the methods names.
        The values are the output of the detailed_evaluate() from Insertion and Deletion.
        (Please refer to this method for further details).
    title
        Title of the graphic.
    filepath
        Path the file will be saved at. If None, the function will call plt.show().
    methods_colors
        Either a dictionary of colors (key are method name, value are matplotlib
        supported color) or string of a matplotlib cmap (e.g 'Set3', 'jet'...).
    """
    # pylint: disable=invalid-name

    methods = list(detailed_scores.keys())

    # set methods colors
    if not isinstance(methods_colors, dict):
        # either None or string
        if methods_colors is None:
            # default cmap
            cmap = matplotlib.cm.get_cmap("Set3")
        else:
            # methods_color is a string linking to a cmap
            cmap = matplotlib.cm.get_cmap(methods_colors)

        methods_colors = {methods[i]: cmap((i + 1) / len(methods))
                          for i in range(len(methods))}

    _, ax = plt.subplots(figsize=(10, 5))
    for method, method_scores in detailed_scores.items():
        ax.plot(
            method_scores.keys(),
            method_scores.values(),
            label=method,
            color=methods_colors[method],
            linewidth=2.0,
        )
    ax.set_title(title)
    ax.legend(title="attribution methods", bbox_to_anchor=(1.05, 1))
    ax.set_xlabel("number of features perturbed")
    ax.set_ylabel("score")
    plt.tight_layout()

    if filepath is not None:
        plt.savefig(filepath)
    else:
        plt.show()



"""
Losses used for feature visualizations
"""

import tensorflow as tf


@tf.function
def cosine_similarity(tensor_a: tf.Tensor, tensor_b: tf.Tensor) -> tf.Tensor:
    """
    Return the cosine similarity for batchs of vectors passed.

    Parameters
    ----------
    tensor_a
        Batch of N tensors.
    tensor_b
        Batch of N tensors.

    Returns
    -------
    cosine_similarity
        The cosine similarity for each pairs of tensors : <x, y> / (|x|+|y|)
    """
    axis_to_norm = range(1, len(tf.shape(tensor_a)))

    tensor_a = tf.nn.l2_normalize(tensor_a, axis=axis_to_norm)
    tensor_b = tf.nn.l2_normalize(tensor_b, axis=axis_to_norm)

    return tf.reduce_sum(tensor_a * tensor_b, axis=axis_to_norm)


@tf.function
def dot_cossim(tensor_a: tf.Tensor, tensor_b: tf.Tensor, cossim_pow: float = 2.0) -> tf.Tensor:
    """
    Return the product of the cosine similarity and the dot product for batchs of vectors passed.
    This original looking loss was proposed by the authors of lucid and seeks to both optimise
    the direction with cosine similarity, but at the same time exaggerate the feature (caricature)
    with the dot product.

    source: https://github.com/tensorflow/lucid/issues/116

    Parameters
    ----------
    tensor_a
        Batch of N tensors.
    tensor_b
        Batch of N tensors.
    cossim_pow
        Power of the cosine similarity, higher value encourage the objective to care more about
        the angle of the activations.

    Returns
    -------
    dot_cossim_value
        The product of the cosine similarity and the dot product for each pairs of tensors.
    """

    cosim = tf.maximum(cosine_similarity(tensor_a, tensor_b), 1e-1) ** cossim_pow
    dot = tf.reduce_sum(tensor_a * tensor_b)

    return dot * cosim



"""
Ensure we can use the operator functionality on various models
"""

import numpy as np
import tensorflow as tf

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, GradCAM, Occlusion, Rise, GuidedBackprop, DeconvNet,
                                  GradCAMPP, Lime, KernelShap, SobolAttributionMethod,
                                  HsicAttributionMethod)
from xplique.commons.operators_operations import semantic_segmentation_operator
from ..utils import generate_data, almost_equal


def default_methods(model, operator):
    return [
        Saliency(model, operator=operator),
        GradientInput(model, operator=operator),
        SmoothGrad(model, operator=operator),
        VarGrad(model, operator=operator),
        SquareGrad(model, operator=operator),
        IntegratedGradients(model, operator=operator),
        Occlusion(model, operator=operator),
        Rise(model, operator=operator, nb_samples=2),
        GuidedBackprop(model, operator=operator),
        DeconvNet(model, operator=operator),
        SobolAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
        HsicAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
    ]


def get_segmentation_model(input_shape=(20, 20, 1)):
    model = tf.keras.Sequential([
        tf.keras.layers.Input(input_shape),
    ])
    model.compile()
    return model


def test_segmentation_operator():
    input_shape = (20, 20, 3)
    segmentation_model = get_segmentation_model(input_shape)

    x, _ = generate_data(input_shape, 10, 10)
    y, _ = generate_data(input_shape, 10, 10)

    methods = default_methods(segmentation_model, semantic_segmentation_operator)
    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]


def test_segmentation_operator_computation():
    image = [[[[0, 0, 1.0],
               [0, 0, 1.0],
               [1.0, 1.0, 1.0],],
              [[0, 0, 0],
               [0, 1.0, 0],
               [1.0, 1.0, 1.0],],
              [[0, 0, 0],
               [0, 0, 0],
               [1.0, 0, 0],],]]
    image = tf.transpose(tf.convert_to_tensor(image, tf.float32), perm=[0, 2, 3, 1])
    
    model = lambda x: tf.concat([x, tf.expand_dims(tf.reduce_mean(x, axis=-1), axis=-1)], axis=-1)

    target_1 = [[[[0, 0, 1.0],
                  [0, 0, 1.0],
                  [1.0, 1.0, 1.0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],]]
    target_1 = tf.transpose(tf.convert_to_tensor(target_1, tf.float32), perm=[0, 2, 3, 1])

    target_2 = [[[[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [0, 0, 0],],
                 [[0, 0, 0],
                  [0, 0, 0],
                  [1.0, 1.0, 1.0],],]]
    target_2 = tf.transpose(tf.convert_to_tensor(target_2, tf.float32), perm=[0, 2, 3, 1])

    scores = model(np.array(image)) * target_2

    score_1 = semantic_segmentation_operator(model, image, np.array(target_1))
    score_2 = semantic_segmentation_operator(model, np.array(image), target_2)
    
    assert almost_equal(score_1, 1.0)
    assert almost_equal(score_2, (7.0 / 3) / 3)




"""
Collections of exceptions used in the library.
"""


class InvalidModelException(Exception):
    """Exception raised when the model is not supported by the library."""


def no_gradients_available():
    """Exception raised when no gradients are available for the specified model."""
    raise InvalidModelException("No gradients are available for the specified "
                          "model. Make sure it is a Keras or Tensorflow model.")


class InvalidOperatorException(Exception):
    """Exception raised when the operator is not supported by the library."""


def raise_invalid_operator():
    """Exception raised when the operator is not supported by the library."""
    raise InvalidOperatorException("The operator must be a function 'g(f,x,y) -> R' that take the"
                                   "model (f), the inputs (x), the label (y) and return a scalar.")



import numpy as np
import tensorflow as tf

from xplique.attributions import GuidedBackprop
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output shape must be the same as the input shape"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 100)
        model = generate_model(input_shape, nb_labels)

        method = GuidedBackprop(model, -2)
        explanations = method.explain(x, y)

        assert x.shape == explanations.shape


def test_guided_mechanism():
    """Ensure we have a proper guided relu instead of relu"""

    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Activation(tf.nn.relu, input_shape=(4,)))
    model.add(tf.keras.layers.Lambda(lambda x: -x ** 2))

    x = tf.constant(np.expand_dims([5.0, 0.0, -5.0, 10.0], axis=0))

    with tf.GradientTape() as tape:
        tape.watch(x)
        y = model(x)
    normal_grads = tape.gradient(y, x).numpy()[0]

    # dy / dx = {-2x for x > 0, else 0 }
    assert almost_equal(normal_grads, np.array([-10.0, 0.0, 0.0, -20.0]), 1e-5)

    guided_model = GuidedBackprop(model).model

    with tf.GradientTape() as tape:
        tape.watch(x)
        y = guided_model(x)
    guided_grads = tape.gradient(y, x).numpy()[0]

    # dy / dx = { 0 }
    assert almost_equal(guided_grads, np.array([0.0, 0.0, 0.0, 0.0]), 1e-5)

    # ensure we didn't change the mechanism of the original model
    with tf.GradientTape() as tape:
        tape.watch(x)
        y = model(x)
    normal_grads_2 = tape.gradient(y, x).numpy()[0]

    assert almost_equal(normal_grads, normal_grads_2, 1e-5)



"""
Module related to abstract explainer
"""

from abc import ABC, abstractmethod
import warnings

import tensorflow as tf
import numpy as np

from ..types import Callable, Dict, Tuple, Union, Optional, OperatorSignature
from ..commons import Tasks
from ..commons import (find_layer, tensor_sanitize, get_inference_function,
                      get_gradient_functions, no_gradients_available)
from ..wrappers import TorchWrapper


def sanitize_input_output(explanation_method: Callable):
    """
    Wrap a method explanation function to ensure tf.Tensor as inputs,
    and as output

    explanation_method
        Function to wrap, should return an tf.tensor.
    """
    def sanitize(self, inputs: Union[tf.data.Dataset, tf.Tensor, np.array],
                 targets: Optional[Union[tf.Tensor, np.array]],
                 *args):
        # ensure we have tf.tensor
        inputs, targets = tensor_sanitize(inputs, targets)
        # then enter the explanation function
        return explanation_method(self, inputs, targets, *args)

    return sanitize


class BlackBoxExplainer(ABC):
    """
    Base class for Black-Box explainers.

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    batch_size
        Number of pertubed samples to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    """

    # in order to avoid re-tracing at each tf.function call,
    # share the reconfigured models between the methods if possible
    _cache_models: Dict[Tuple[int, int], tf.keras.Model] = {}

    def __init__(self, model: Callable, batch_size: Optional[int] = 64,
                operator: Optional[Union[Tasks, str, OperatorSignature]] = None):

        if isinstance(model, TorchWrapper):
            self.model = model
        elif isinstance(model, tf.keras.Model):
            model_key = (id(model.input), id(model.output))
            if model_key not in BlackBoxExplainer._cache_models:
                BlackBoxExplainer._cache_models[model_key] = model
            self.model = BlackBoxExplainer._cache_models[model_key]
        else:
            self.model = model

        self.batch_size = batch_size

        # define the inference function according to the model type
        self.inference_function, self.batch_inference_function = \
            get_inference_function(model, operator)

        # black box method don't have access to the model's gradients
        self.gradient = no_gradients_available
        self.batch_gradient = no_gradients_available


    @abstractmethod
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.array],
                targets: Optional[Union[tf.Tensor, np.array]] = None) -> tf.Tensor:
        """
        Compute the explanations of the given inputs.
        Accept Tensor, numpy array or tf.data.Dataset (in that case targets is None)

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Explanation generated by the method.
        """
        raise NotImplementedError()

    def __call__(self,
                 inputs: tf.Tensor,
                 labels: tf.Tensor) -> tf.Tensor:
        """Explain alias"""
        return self.explain(inputs, labels)


class WhiteBoxExplainer(BlackBoxExplainer, ABC):
    """
    Base class for White-Box explainers.

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Operator to use to compute the explanation, if None use standard predictions.
    """

    def __init__(self,
                model: tf.keras.Model,
                output_layer: Optional[Union[str, int]] = None,
                batch_size: Optional[int] = 64,
                operator: Optional[OperatorSignature] = None):

        super().__init__(model, batch_size, operator)

        if output_layer is not None:
            # reconfigure the model (e.g skip softmax to target logits)
            target_layer = find_layer(model, output_layer)
            model = tf.keras.Model(model.input, target_layer.output)

            # sanity check, output layer before softmax
            try:
                if target_layer.activation.__name__ == tf.keras.activations.softmax.__name__:
                    warnings.warn("Output is after softmax, it is recommended to "
                                  "use the layer before.")
            except AttributeError:
                pass

        # check and get gradient function from model and operator
        self.gradient, self.batch_gradient = get_gradient_functions(model, operator)



"""
Ensure we can use the operator functionnality on various models
"""

import pytest

import xplique
from xplique.commons.operators_operations import check_operator, Tasks, get_operator
from xplique.commons.operators import (predictions_operator, regression_operator,
                                       semantic_segmentation_operator, object_detection_operator)
from xplique.commons.exceptions import InvalidOperatorException


def test_check_operator():
    # ensure that the check operator detects non-operator

    # operator must have at least 3 arguments
    function_with_2_arguments = lambda x,y: 0

    # operator must be Callable
    not_a_function = [1, 2, 3]

    for operator in [function_with_2_arguments, not_a_function]:
        try:
            check_operator(operator)
            assert False
        except InvalidOperatorException:
            pass


def test_get_operator():
    possible_tasks = ["classification", "regression", "semantic segmentation", "object detection",
                      "object detection box position", "object detection box proba",
                      "object detection box class"]

    tasks_name = [task.name for task in Tasks]
    assert tasks_name.sort() == possible_tasks.sort()

    # get by enum
    assert get_operator(Tasks.CLASSIFICATION) is predictions_operator
    assert get_operator(Tasks.REGRESSION) is predictions_operator  # TODO, change when there is a real regression operator
    assert get_operator(Tasks.OBJECT_DETECTION) is object_detection_operator
    assert get_operator(Tasks.SEMANTIC_SEGMENTATION) is semantic_segmentation_operator

    # get by string
    assert get_operator("classification") is predictions_operator
    assert get_operator("regression") is predictions_operator  # TODO, change when there is a real regression operator
    assert get_operator("object detection") is object_detection_operator
    assert get_operator("semantic segmentation") is semantic_segmentation_operator

    # assert a not valid string does not work
    with pytest.raises(AssertionError):
        get_operator("random")

    # operator must have at least 3 arguments
    function_with_2_arguments = lambda x,y: 0

    # operator must be Callable
    not_a_function = [1, 2, 3]

    for operator in [function_with_2_arguments, not_a_function]:
        try:
            get_operator(operator)
        except InvalidOperatorException:
            pass


def test_proposed_operators():
    # ensure all proposed operators are operators
    for operator in [task.value for task in Tasks]:
        check_operator(operator)

def test_enum_shortcut():
    # ensure all proposed operators are operators
    for operator in [task.value for task in xplique.Tasks]:
        check_operator(operator)



import numpy as np
import tensorflow as tf

from xplique.concepts import Tcav


def test_tcav_computation():
    """For a given simple example, ensure that the computation of tcav works"""

    # we create a toy example for a function f(x1, x2) -> y1, y2
    # and y1 = x1 > 0.5, y2 = x2 > 0.5
    x = np.random.random((1000, 2))
    y = np.array(x > 0.5, np.float32)

    model = tf.keras.Sequential([
        tf.keras.layers.Input(2),
        tf.keras.layers.Lambda(lambda x: tf.cast(x > 0.5, tf.float32)),
        tf.keras.layers.Dense(2, activation='linear')
    ])
    model.compile(loss='mse', optimizer=tf.keras.optimizers.Adam(0.1))
    model.fit(x, y, batch_size=32, epochs=1)

    # by forcing the model to learn this mapping, we should have the following
    # score for the cav :
    tcav_maker = Tcav(model, 1)

    # cav [1.0, 0.0] relative to y1 = 1
    # cav [0.0, 1.0] relative to y1 = 0
    # cav [1.0, 0.0] relative to y2 = 0
    # cav [0.0, 1.0] relative to y2 = 1
    tcav_x1_y1 = tcav_maker(x, 0, [1.0, 0.0])
    tcav_x2_y1 = tcav_maker(x, 0, [0.0, 1.0])
    tcav_x1_y2 = tcav_maker(x, 1, [1.0, 0.0])
    tcav_x2_y2 = tcav_maker(x, 1, [0.0, 1.0])

    epsilon = 0.01
    assert np.abs(tcav_x1_y1 - 1.0) < epsilon
    assert np.abs(tcav_x1_y2) < epsilon
    assert np.abs(tcav_x2_y1) < epsilon
    assert np.abs(tcav_x2_y2 - 1.0) < epsilon



from xplique.features_visualizations.preconditioning import fft_image, \
    fft_to_rgb, get_fft_scale

def test_shape():
    """Ensure the shape of the image generated is correct"""

    input_shapes = [(2, 8, 8, 3), (2, 64, 64, 1), (2, 256, 256, 3), (2, 512, 512, 1)]

    for input_shape in input_shapes:
        n, w, h, c = input_shape
        buffer = fft_image(input_shape)
        scaler = get_fft_scale(w, h)
        img = fft_to_rgb(input_shape, buffer, scaler)

        # image generated from inverse fourier transformation should have
        # the desired shape
        assert img.shape == input_shape



"""
Module related to the Testing of CAVs
"""

import tensorflow as tf
import numpy as np

from ..commons import find_layer, batch_tensor
from ..types import Union, Optional


class Tcav:
    """
    Used to Test a Concept Activation Vector, using the sign of the directional derivative of a
    concept vector relative to a class.

    Ref. Kim & al., Interpretability Beyond Feature Attribution: Quantitative Testing with Concept
    Activation Vectors (TCAV) (2018).
    https://arxiv.org/abs/1711.11279

    Parameters
    ----------
    model
        Model to extract concept from.
    target_layer
        Index of the target layer or name of the layer.
    batch_size
        Batch size during the predictions.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 target_layer: Union[str, int],
                 batch_size: Optional[int] = 64):
        self.model = model
        self.batch_size = batch_size

        # configure model bottleneck
        target_layer = find_layer(model, target_layer)
        self.multi_head = tf.keras.Model(model.input, [target_layer.output, model.output])

    def score(self,
              inputs: tf.Tensor,
              label: int,
              cav: tf.Tensor) -> float:
        """
        Compute and return the TCAV score of the CAV associated to class tested.

        Parameters
        ----------
        inputs
            Input sample on which to test the influence of the concept.
        label
            Index of the class to test.
        cav
            Concept Activation Vector, see CAV module.

        Returns
        -------
        tcav
            Percentage of sample for which increasing the concept has a positive impact on the
            class logit.
        """

        directional_derivatives = None
        label = tf.cast(label, tf.int32)
        cav = tf.cast(cav, tf.float32)

        batch_size = self.batch_size or len(inputs)

        for x_batch in batch_tensor(inputs, batch_size):
            batch_dd = Tcav.directional_derivative(self.multi_head,
                                                   x_batch, label,
                                                   cav)
            directional_derivatives = batch_dd if directional_derivatives is None else \
                tf.concat([directional_derivatives, batch_dd], axis=0)

        # tcav is the number of positive directional derivatives
        tcav = np.mean(directional_derivatives > 0.0)

        return tcav

    __call__ = score

    @staticmethod
    @tf.function
    def directional_derivative(multi_head_model: tf.keras.Model,
                               inputs: tf.Tensor,
                               label: int,
                               cav: tf.Tensor) -> tf.Tensor:
        """
        Compute the gradient of the label relative to the activations of the CAV layer.

        Parameters
        ----------
        multi_head_model
            Model reconfigured, first output is the activations of the CAV layer, and the second
            output is the prediction layer.
        inputs
            Input sample on which to test the influence of the concept.
        label
            Index of the class to test.
        cav
            Concept Activation Vector, same shape as the activations output.

        Returns
        -------
        directional_derivative
            Directional derivative values of each samples.
        """
        with tf.GradientTape() as tape:
            tape.watch(inputs)

            activations, y_pred = multi_head_model(inputs)
            score = y_pred[:, label]

        gradients = tape.gradient(score, activations)

        # compute the directional derivatives in terms of partial derivatives
        axis_to_reduce = tf.range(1, tf.rank(gradients))
        directional_derivative = tf.reduce_sum(gradients * cav, axis=axis_to_reduce)

        return directional_derivative



"""
Images preconditionners

Adaptation of the original Lucid library :
https://github.com/tensorflow/lucid/blob/master/lucid/optvis/param/color.py
Credit is due to the original Lucid authors.
"""


import numpy as np
import tensorflow as tf

from ..types import Tuple, Union, Callable


IMAGENET_SPECTRUM_URL = "https://storage.googleapis.com/serrelab/loupe/"\
                        "spectrums/imagenet_decorrelated.npy"

imagenet_color_correlation = tf.cast(
      [[0.56282854, 0.58447580, 0.58447580],
       [0.19482528, 0.00000000,-0.19482528],
       [0.04329450,-0.10823626, 0.06494176]], tf.float32
)


def recorrelate_colors(images: tf.Tensor) -> tf.Tensor:
    """
    Map uncorrelated colors to 'normal colors' by using empirical color
    correlation matrix of ImageNet (see https://distill.pub/2017/feature-visualization/)

    Parameters
    ----------
    images
        Input samples , with N number of samples, W & H the sample dimensions,
        and C the number of channels.

    Returns
    -------
    images
        Images recorrelated.
    """
    images_flat = tf.reshape(images, [-1, 3])
    images_flat = tf.matmul(images_flat, imagenet_color_correlation)
    return tf.reshape(images_flat, tf.shape(images))


def to_valid_rgb(images: tf.Tensor,
                 normalizer: Union[str, Callable] = 'sigmoid',
                 values_range: Tuple[float, float] = (0, 1)) -> tf.Tensor:
    """
    Apply transformations to map tensors to valid rgb images.


    Parameters
    ----------
    images
        Input samples, with N number of samples, W & H the sample dimensions,
        and C the number of channels.
    normalizer
        Transformation to apply to map pixels in the range [0, 1]. Either 'clip' or 'sigmoid'.
    values_range
        Range of values of the inputs that will be provided to the model, e.g (0, 1) or (-1, 1).

    Returns
    -------
    images
        Images after correction
    """
    images = recorrelate_colors(images)

    if normalizer == 'sigmoid':
        images = tf.nn.sigmoid(images)
    elif normalizer == 'clip':
        images = tf.clip_by_value(images, values_range[0], values_range[1])
    else:
        images = normalizer(images)

    # rescale according to value range
    images = images - tf.reduce_min(images, (1, 2, 3), keepdims=True)
    images = images / tf.reduce_max(images, (1, 2, 3), keepdims=True)
    images *= values_range[1] - values_range[0]
    images += values_range[0]

    return images


def fft_2d_freq(width: int, height: int) -> np.ndarray:
    """
    Return the fft samples frequencies for a given width/height.
    As we deal with real values (pixels), the Discrete Fourier Transform is
    Hermitian symmetric, tensorflow's reverse operation requires only
    the unique components (width, height//2+1).

    Parameters
    ----------
    width
        Width of the image.
    height
        Height of the image.

    Returns
    -------
    frequencies
        Array containing the samples frequency bin centers in cycles per pixels
    """
    freq_y = np.fft.fftfreq(height)[:, np.newaxis]

    cut_off = int(width % 2 == 1)
    freq_x = np.fft.fftfreq(width)[:width//2+1+cut_off]

    return np.sqrt(freq_x**2 + freq_y**2)


def get_fft_scale(width: int, height: int, decay_power: float = 1.0) -> tf.Tensor:
    """
    Generate 'scaler' to normalize spectrum energy. Also scale the energy by the
    dimensions to use similar learning rate regardless of image size.
    adaptation of : https://github.com/tensorflow/lucid/blob/master/lucid/optvis/param/spatial.py
    #L73

    Parameters
    ----------
    width
        Width of the image.
    height
        Height of the image.
    decay_power
        Control the allowed energy of the high frequency, a high value
        suppresses high frequencies.

    Returns
    -------
    fft_scale
        Scale factor of the fft spectrum
    """
    frequencies = fft_2d_freq(width, height)
    fft_scale = 1.0 / np.maximum(frequencies, 1.0 / max(width, height)) ** decay_power
    fft_scale = fft_scale * np.sqrt(width * height)

    return tf.convert_to_tensor(fft_scale, dtype=tf.complex64)


def fft_to_rgb(shape: Tuple, buffer: tf.Tensor, fft_scale: tf.Tensor) -> tf.Tensor:
    """
    Convert a fft buffer into images.

    Parameters
    ----------
    shape
        Shape of the images with N number of samples, W & H the sample
        dimensions, and C the number of channels.
    buffer
        Image buffer in the fourier basis.
    fft_scale
        Scale factor of the fft spectrum

    Returns
    -------
    images
        Images in the 'pixels' basis.
    """
    batch, width, height, channels = shape
    spectrum = tf.complex(buffer[0], buffer[1]) * fft_scale

    image = tf.signal.irfft2d(spectrum)
    image = tf.transpose(image, (0, 2, 3, 1))
    image = image[:batch, :width, :height, :channels]

    return image / 4.0


def fft_image(shape: Tuple, std: float = 0.01) -> tf.Tensor:
    """
    Generate the preconditioned image buffer

    Parameters
    ----------
    shape
        Shape of the images with N number of samples, W & H the sample
        dimensions, and C the number of channels.
    std
        Standard deviation of the normal for the buffer initialization
    Returns
    -------
    buffer
        Image buffer in the fourier basis.
    """
    batch, width, height, channels = shape
    frequencies = fft_2d_freq(width, height)

    buffer = tf.random.normal((2, batch, channels)+frequencies.shape,
                              stddev=std)

    return buffer


def init_maco_buffer(image_shape, std=1.0):
    """
    Initialize the buffer for the MACO algorithm.

    Parameters
    ----------
    image_shape
        Shape of the images with N number of samples, W & H the sample
        dimensions, and C the number of channels.
    std
        Standard deviation of the normal for the buffer initialization

    Returns
    -------
    magnitude
        Magnitude of the spectrum
    phase
        Phase of the spectrum
    """
    spectrum_shape = (image_shape[0], image_shape[1]//2+1)

    # init randomly the phase and load the constrained spectrum (average spectrum)
    phase = np.random.normal(size=(3, *spectrum_shape), scale=std).astype(np.float32)

    magnitude_path = tf.keras.utils.get_file("spectrum_decorrelated.npy",
                                             IMAGENET_SPECTRUM_URL,
                                             cache_subdir="spectrums")
    magnitude = np.load(magnitude_path)
    magnitude = tf.image.resize(np.moveaxis(magnitude, 0, -1), spectrum_shape).numpy()
    magnitude = np.moveaxis(magnitude, -1, 0)

    return tf.cast(magnitude, tf.float32), tf.cast(phase, tf.float32)

@tf.function
def maco_image_parametrization(magnitude, phase, values_range):
    """
    Generate the image from the magnitude and phase using MaCo method.

    Parameters
    ----------
    magnitude
        Magnitude of the spectrum
    phase
        Phase of the spectrum
    values_range
        Range of the values of the image

    Returns
    -------
    img
        Image in the 'pixels' basis.
    """
    phase = phase - tf.reduce_mean(phase)
    phase = phase / (tf.math.reduce_std(phase) + 1e-5)

    buffer = tf.complex(tf.cos(phase) * magnitude, tf.sin(phase) * magnitude)
    img = tf.signal.irfft2d(buffer)
    img = tf.transpose(img, [1,2,0])

    img = img - tf.reduce_mean(img)
    img = img / (tf.math.reduce_std(img) + 1e-5)

    img = recorrelate_colors(img)
    img = tf.nn.sigmoid(img)

    img = img * (values_range[1] - values_range[0]) + values_range[0]

    return img



"""
Pretty plots option for explanations
"""

import matplotlib
from matplotlib import pyplot as plt
import numpy as np
import tensorflow as tf

from ..types import Union, Dict, List, Any


def _arrange_subplots(
        explanations: Dict[str, Union[np.array, tf.Tensor]],
) -> Dict[str, Any]:
    """
    Determine how to arrange subplots so that the final output have coherent proportions.

    Parameters
    ----------
    explanations
        dictionary of explanations

    Returns
    -------
    subplot_kwargs
        Dictionary of arguments for the plt.subplots() method.

    """
    # initialize sizes and shapes
    nb_subplots = len(explanations)
    expl_shape = list(explanations.values())[0].shape
    # matplotlib dimension are flipped compare to usual numpy of pandas
    get_size = lambda x, y: (int(y * (1 + expl_shape[0] / 5)), int(x * (1 + expl_shape[1] / 5)))

    found_arrange = False
    # iterate to found arrangement
    while not found_arrange:
        for i in range(1, nb_subplots + 1):
            if nb_subplots % i == 0:
                plot_size = get_size(i, nb_subplots / i)
                if plot_size[0] <= plot_size[1] <= 2 * plot_size[0]:
                    found_arrange = True
                    nrows = i
                    ncols = int(nb_subplots / i)
                    break
        nb_subplots += 1
    return {"nrows": nrows, "ncols": ncols, "figsize": plot_size}


def _show_heatmap(
        axe: matplotlib.axes.Axes,
        explanations: Union[np.array, tf.Tensor],
        features: List[str],
        title: str = "",
        cmap: str = "coolwarm",
        **plot_kwargs
) -> matplotlib.image.AxesImage:
    """
    Display a heatmap representing the attributions for timeseries.

    Parameters
    ----------
    explanations
        Attributions, they are numpy arrays or tensorflow tensors.
        (With features * timesteps format).
    features
        List of features names, should match the first dimension of explanations
    title
        Title of the plot.
    cmap
        Matplotlib color map to apply.
    plot_kwargs
        Additional parameters passed to `plt.imshow()`.

    Returns
    -------
    image
        output of ax.imshow().
    """
    image = axe.imshow(np.array(explanations).transpose(), cmap=cmap, **plot_kwargs)

    axe .set_title(title, fontsize=12)

    axe .set_xlabel("time-steps", fontsize=10)
    time_steps = list(range(-explanations.shape[0], 0))
    axe .set_xticks(np.arange(len(time_steps)))
    axe .set_xticklabels(time_steps, fontsize=5)

    axe .set_ylabel("features", fontsize=10)
    axe .set_yticks(np.arange(len(features)))
    axe .set_yticklabels(features, fontsize=8)

    return image


def plot_attributions(
        explanations: Union[Dict[str, Union[np.array, tf.Tensor]], np.array, tf.Tensor],
        features: List[str],
        title: str = "",
        filepath: str = None,
        cmap: str = "coolwarm",
        colorbar: bool = False,
        **plot_kwargs,
):
    """
    Display a heatmap representing the attributions for timeseries,
    if it is called with a dictionary of explanations,
    it will make subplots and show an heatmap for all elements of the dictionary.

    Parameters
    ----------
    explanations
        Can either be a dictionary of explanations or one explanation.
        The explanations are numpy arrays or tensorflow tensors. (With features * timesteps format).
    features
        List of features names, should match the first dimension of explanations
    title
        Title of the plot.
    filepath
        Path the file will be saved at.
        If None, the function will call plt.show()
    cmap
        Matplotlib color map to apply.
    colorbar
        If the color bar should be shown.
    plot_kwargs
        Additional parameters passed to `plt.imshow()`.
    """
    if isinstance(explanations, dict):
        # find the right arrangement
        subplot_kwargs = _arrange_subplots(explanations)
        nrows, ncols = subplot_kwargs["nrows"], subplot_kwargs["ncols"]
        fig, axes = plt.subplots(**subplot_kwargs)

        # plot multiple heatmaps
        for i, (method, explanation) in enumerate(explanations.items()):
            if nrows == 1 or ncols == 1:
                axe = axes[i]
            else:
                axe = axes[i % nrows, i // nrows]
            image = _show_heatmap(
                axe,
                explanation,
                features,
                title=method,
                cmap=cmap,
                **plot_kwargs
            )
        fig.suptitle(title, fontsize=16)
        plt.subplots_adjust(wspace=0.4)
        if colorbar:
            fig.colorbar(image, ax=axes.ravel().tolist())
    else:
        # plot a single heatmap
        fig, axe = plt.subplots(
            figsize=(1 + explanations.shape[1] / 5, 1 + explanations.shape[0] / 5)
        )
        image = _show_heatmap(axe, explanations, features, title=title, cmap=cmap, **plot_kwargs)
        if colorbar:
            fig.colorbar(image, ax=axe)
        fig.tight_layout()

    if filepath is not None:
        plt.savefig(filepath)
    else:
        plt.show()



"""
Module related to Grad-CAM++ method
"""

import tensorflow as tf

from .grad_cam import GradCAM


class GradCAMPP(GradCAM):
    """
    Used to compute the Grad-CAM++ visualization method.

    Only for Convolutional Networks.

    Ref. Chattopadhyay & al., Grad-CAM++: Improved Visual Explanations for Deep Convolutional
    Networks (2017).
    https://arxiv.org/abs/1710.11063

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    conv_layer
        Layer to target for Grad-CAM++ algorithm.
        If an int is provided it will be interpreted as a layer index.
        If a string is provided it will look for the layer name.
    """

    # Avoid zero division during procedure. (the value is not important, as if the denominator is
    # zero, then the nominator will also be zero).
    EPSILON = tf.constant(1e-4)

    @staticmethod
    @tf.function
    def _compute_weights(feature_maps_gradients: tf.Tensor,
                         feature_maps: tf.Tensor) -> tf.Tensor:
        """
        Compute the weights according to Grad-CAM++ procedure.

        Parameters
        ----------
        feature_maps_gradients
            Gradients for the target convolution layer.
        feature_maps
            Activations for the target convolution layer.

        Returns
        -------
        weights
            Weights for each feature maps.
        """
        feature_maps_gradients_square = tf.pow(feature_maps_gradients, 2)
        feature_maps_gradients_cube = tf.pow(feature_maps_gradients, 3)

        feature_map_avg = tf.reduce_mean(feature_maps, axis=(1, 2), keepdims=True)

        nominator = feature_maps_gradients_square
        denominator = 2.0 * feature_maps_gradients_square + \
                      feature_maps_gradients_cube * feature_map_avg
        denominator += tf.cast(denominator == 0, tf.float32) * GradCAMPP.EPSILON

        feature_map_alphas = nominator / denominator * tf.nn.relu(feature_maps_gradients)
        weights = tf.reduce_mean(feature_map_alphas, axis=(1, 2))
        weights = tf.reshape(weights,
                             (weights.shape[0], 1, 1,
                              weights.shape[-1]))

        return weights



"""
Hsic Attribution Method explainer
"""

from ...types import Callable, Union, Optional, OperatorSignature
from ...commons import Tasks
from .gsa_attribution_method import GSABaseAttributionMethod
from .samplers import Sampler, TFSobolSequence
from .hsic_estimators import (
    BinaryEstimator,
    HsicEstimator,
)


class HsicAttributionMethod(GSABaseAttributionMethod):
    """
    HSIC Attribution Method.
    Compute the dependance of each input dimension wrt the output using Hilbert-Schmidt Independance
    Criterion, a perturbation function on a grid and an adapted sampling as described in
    the original paper.

    Ref. Novello, Fel, Vigouroux, Making Sense of Dependance: Efficient Black-box Explanations
    Using Dependence Measure, https://arxiv.org/abs/2206.06219

    Parameters
    ----------
    model
        Model used for computing explanations.
    grid_size
        Cut the image in a grid of (grid_size, grid_size) to estimate an indice per cell.
    nb_design
        Number of design for the sampler.
    sampler
        Sampler used to generate the (quasi-)monte carlo samples, LHS or QMC.
        For more option, see the sampler module. Note that the original paper uses LHS but here
        the default sampler is TFSobolSequence as LHS requires scipy 1.7.0.
    estimator
        Estimator used to compute the HSIC score.
    perturbation_function
        Function to call to apply the perturbation on the input. Can also be string in
        'inpainting', 'blur'.
    batch_size
        Batch size to use for the forwards.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    """

    def __init__(
        self,
        model,
        grid_size: int = 8,
        nb_design: int = 500,
        sampler: Optional[Sampler] = None,
        estimator: Optional[HsicEstimator] = None,
        perturbation_function: Optional[Union[Callable, str]] = "inpainting",
        batch_size=256,
        operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
    ):

        sampler = sampler if sampler is not None else TFSobolSequence(binary=True)
        estimator = (
            estimator if estimator is not None else BinaryEstimator(output_kernel="rbf")
        )

        assert isinstance(sampler, Sampler), "The sampler must be a valid Sampler."
        assert isinstance(
            estimator, HsicEstimator
        ), "The estimator must be a valid HsicEstimator."

        if isinstance(estimator, BinaryEstimator):
            assert sampler.binary, "The sampler must be binary for BinaryEstimator."

        super().__init__(model = model, operator = operator,
                         sampler = sampler, estimator = estimator,
                         grid_size = grid_size, nb_design = nb_design,
                         perturbation_function = perturbation_function, batch_size = batch_size,
        )



"""
Module related to Integrated Gradients method
"""

import tensorflow as tf
import numpy as np

from .base import WhiteBoxExplainer, sanitize_input_output
from ..commons import repeat_labels, batch_tensor, Tasks
from ..types import Tuple, Union, Optional, OperatorSignature


class IntegratedGradients(WhiteBoxExplainer):
    """
    Used to compute the Integrated Gradients, by cumulating the gradients along a path from a
    baseline to the desired point.

    Ref. Sundararajan & al., Axiomatic Attribution for Deep Networks (2017).
    http://proceedings.mlr.press/v70/sundararajan17a/sundararajan17a.pdf

    Notes
    -----
    In order to approximate from a finite number of steps, the implementation here use the
    trapezoidal rule and not a Riemann sum (see the paper below for a comparison of the results).
    Ref. Computing Linear Restrictions of Neural Networks (2019).
    https://arxiv.org/abs/1908.06214

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    steps
        Number of points to interpolate between the baseline and the desired point.
    baseline_value
        Scalar used to create the the baseline point.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 output_layer: Optional[Union[str, int]] = None,
                 batch_size: Optional[int] = 32,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 steps: int = 50,
                 baseline_value: float = .0):
        super().__init__(model, output_layer, batch_size, operator)
        self.steps = steps
        self.baseline_value = baseline_value

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute Integrated Gradients for a batch of samples.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Integrated gradients, same shape as the inputs.
        """
        integrated_gradients = None
        batch_size = self.batch_size or len(inputs)
        baseline = IntegratedGradients._get_baseline((*inputs.shape[1:],),
                                                     self.baseline_value)

        for x_batch, y_batch in batch_tensor((inputs, targets),
                                             max(batch_size // self.steps, 1)):
            # create the paths for every sample (interpolated points from baseline to sample)
            interpolated_inputs = IntegratedGradients._get_interpolated_points(
                x_batch, self.steps, baseline)
            repeated_targets = repeat_labels(y_batch, self.steps)

            # compute the gradient for each paths
            interpolated_gradients = self.batch_gradient(self.model, interpolated_inputs,
                                                         repeated_targets, batch_size)
            interpolated_gradients = tf.reshape(interpolated_gradients,
                                                (-1, self.steps, *interpolated_gradients.shape[1:]))

            # average the gradient using trapezoidal rule
            averaged_gradients = IntegratedGradients._average_gradients(interpolated_gradients)
            batch_integrated_gradients = (x_batch - baseline) * averaged_gradients

            integrated_gradients = batch_integrated_gradients if integrated_gradients is None else \
                tf.concat([integrated_gradients, batch_integrated_gradients], axis=0)

        return integrated_gradients

    @staticmethod
    def _get_baseline(shape: Tuple,
                      baseline_value: float) -> tf.Tensor:
        """
        Create the baseline point using a scalar value to fill the desired shape.

        Parameters
        ----------
        shape
            Desired shape, dimension of one sample.
        baseline_value
            Value defining the baseline state.

        Returns
        -------
        baseline_point
            A baseline point of the specified shape.
        """
        return tf.ones(shape) * baseline_value

    @staticmethod
    @tf.function
    def _get_interpolated_points(inputs: tf.Tensor,
                                 steps: int,
                                 baseline: tf.Tensor) -> tf.Tensor:
        """
        Create a path from baseline to sample for every samples.

        Parameters
        ----------
        inputs
            Input samples to be explained.
        steps
            Number of points to interpolate between the baseline and the desired point.
        baseline
            Baseline point, start of the path.

        Returns
        -------
        interpolated_inputs
            Interpolated path for each inputs, the first dimension correspond to the number of
            samples multiplied by steps.
        """
        alpha = tf.reshape(tf.linspace(0.0, 1.0, steps), (1, -1, *(1,) * len(inputs.shape[1:])))

        interpolated_inputs = tf.expand_dims(inputs, axis=1)
        interpolated_inputs = tf.repeat(interpolated_inputs, repeats=steps, axis=1)
        interpolated_inputs = baseline + alpha * (interpolated_inputs - baseline)

        interpolated_inputs = tf.reshape(interpolated_inputs, (-1, *interpolated_inputs.shape[2:]))

        return interpolated_inputs

    @staticmethod
    @tf.function
    def _average_gradients(gradients: tf.Tensor) -> tf.Tensor:
        """
        Average the gradients obtained along the path using trapezoidal rule.

        Parameters
        ----------
        gradients
            Gradients obtained for each of the steps for each of the samples.

        Returns
        -------
        integrated_gradients
        """
        trapezoidal_gradients = gradients[:, :-1] + gradients[:, 1:]
        averaged_gradients = tf.reduce_mean(trapezoidal_gradients, axis=1) * 0.5

        return averaged_gradients



"""
Custom tensorflow operator for Attributions
"""

from deprecated import deprecated

import tensorflow as tf

from ..types import Callable, Optional
from ..utils_functions.object_detection import _box_iou, _format_objects, _EPSILON


@tf.function
def predictions_operator(model: Callable,
                         inputs: tf.Tensor,
                         targets: tf.Tensor) -> tf.Tensor:
    """
    Compute predictions scores, only for the label class, for a batch of samples.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    scores
        Predictions scores computed, only for the label class.
    """
    scores = tf.reduce_sum(model(inputs) * targets, axis=-1)
    return scores

@tf.function
@deprecated(version="1.0.0", reason="Gradient-based explanations are zeros with this operator.")
def regression_operator(model: Callable,
                        inputs: tf.Tensor,
                        targets: tf.Tensor) -> tf.Tensor:
    """
    Compute the the mean absolute error between model prediction and the target.
    Target should the model prediction on non-perturbed input.
    This operator can be used to compute attributions for all outputs of a regression model.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        Model prediction on non-perturbed inputs.

    Returns
    -------
    scores
        MAE between model prediction and targets.
    """
    scores = tf.reduce_mean(tf.abs(model(inputs) - targets), axis=-1)
    return scores


@tf.function
def semantic_segmentation_operator(model, inputs, targets):
    """
    Explain the class of a zone of interest.

    Parameters
    ----------
    model
        Model used for computing predictions.
        The model outputs should be between 0 and 1, otherwise, applying a softmax is recommended.
    inputs
        Input samples to be explained.
        Expected shape of (n, h, w, c_in), with c_in the number of channels of the input.
    targets
        Tensor, a mask indicating the zone and class to explain.
        It contains the model predictions limited to a certain zone and channel.
        The zone indicates the zone of interest and the channel the class of interest.
        For more detail and examples please refer to the documentation.
        https://deel-ai.github.io/xplique/latest/api/attributions/semantic_segmentation/
        Expected shape of (n, h, w, c_out), with c_out the number of classes.
        `targets` can also be designed to explain the border of a zone of interest.

    Returns
    -------
    scores
        Segmentation scores computed.
    """
    # compute absolute difference between prediction and target on targets zone
    scores = model(inputs) * targets

    # take mean over the zone and channel of interest
    return tf.reduce_sum(scores, axis=(1, 2, 3)) /\
        tf.reduce_sum(tf.cast(tf.not_equal(targets, 0), tf.float32), axis=(1, 2, 3))


@tf.function
def object_detection_operator(model: Callable,
                              inputs: tf.Tensor,
                              targets: tf.Tensor,
                              intersection_score_fn: Optional[Callable]  = _box_iou,
                              include_detection_probability: Optional[bool] = True,
                              include_classification_score: Optional[bool] = True,) -> tf.Tensor:
    """
    Compute the object detection scores for a batch of samples.

    For a given image, there are two possibilities:
        - One box per image is provided: Then, in the case of perturbation-based methods,
        the model makes prediction on the perturbed image and choose the most similar predicted box.
        This similarity is computed following the DRise method.
        In the case of gradient-based methods, the gradient is computed from the same score.
        - Several boxes are provided for one image: In this case, the attributions for each box are
        computed and the mean is taken.

    Therefore, to explain each box separately, the easiest way is to call the attribution method
    with a batch of the same image tiled to match the number of predicted box.
    In this case, inputs and targets shapes should be: (nb_boxes, H, W, C) and (nb_boxes, (5 + nc)).

    This work is a generalization of the following article at any kind of attribution method.
    Ref. Petsiuk & al., Black-box Explanation of Object Detectors via Saliency Maps (2021).
    https://arxiv.org/pdf/2006.03204.pdf

    Parameters
    ----------
    model
        Model used for computing object detection prediction.
        The model should have input and output shapes of (N, H, W, C) and (N, nb_boxes, (4+1+nc)).
        The model should not include the NMS computation,
        it is not differentiable and drastically reduce the number of boxes for the matching.
    inputs
        Batched input samples to be explained. Expected shape (N, H, W, C).
        More information in the documentation.
    targets
        Specify the box are boxes to explain for each input. Preferably, after the NMS.
        It should be of shape (N, (4 + 1 + nc)) or (N, nb_boxes, (4 + 1 + nc)),
        with nc the number of classes,
        N the number of samples in the batch (it should match `inputs`),
        and nb_boxes the number of boxes to explain simultaneously.

        (4 + 1 + nc) means: [boxes_coordinates, proba_detection, one_hots_classifications].

        In the case the nb_boxes dimension is not 1,
        several boxes will be explained at the same time.
        To be more precise, explanations will be computed for each box and the mean is returned.
    intersection_score_fn
        Function that computes the intersection score between two bounding boxes coordinates.
        This function is batched. The default value is `_box_iou` computing IOU scores.
    include_detection_probability
        Boolean encoding if the box objectness (or detection probability)
        should be included in DRise score.
    include_classification_score
        Boolean encoding if the class associated to the box should be included in DRise score.

    Returns
    -------
    scores
        Object detection scores computed following DRise definition:
        intersection_score * proba_detection * classification_similarity
    """
    def batch_loop(args):
        # function to loop on for `tf.map_fn`
        obj, obj_ref = args

        if obj is None or obj.shape[0] == 0:
            return tf.constant(0.0, dtype=inputs.dtype)

        # compute predicted boxes for a given image
        # (nb_box_pred, 4), (nb_box_pred, 1), (nb_box_pred, nb_classes)
        current_boxes, proba_detection, classification = _format_objects(obj)
        size = tf.shape(current_boxes)[0]

        if len(tf.shape(obj_ref)) == 1:
            obj_ref = tf.expand_dims(obj_ref, axis=0)

        # DRise consider the reference objectness to be 1
        # (nb_box_ref, 4), _, (nb_box_ref, nb_classes)
        boxes_refs, _, class_refs = _format_objects(obj_ref)

        # (nb_box_ref, nb_box_pred, 4)
        boxes_refs = tf.repeat(tf.expand_dims(boxes_refs, axis=1), repeats=size, axis=1)

        # (nb_box_ref, nb_box_pred)
        intersection_score = intersection_score_fn(boxes_refs, current_boxes)

        # (nb_box_pred,)
        detection_probability = tf.squeeze(proba_detection, axis=1)

        # set detection probability to 1 if it should be included
        detection_probability = tf.cond(tf.cast(include_detection_probability, tf.bool),
                                        true_fn=lambda: detection_probability,
                                        false_fn=lambda: tf.ones_like(detection_probability))

        # (nb_box_ref, nb_box_pred, nb_classes)
        class_refs = tf.repeat(tf.expand_dims(class_refs, axis=1), repeats=size, axis=1)

        # (nb_box_ref, nb_box_pred)
        classification_score = tf.reduce_sum(class_refs * classification, axis=-1) \
                / (tf.norm(classification, axis=-1) * tf.norm(class_refs, axis=-1)+ _EPSILON)

        # set classification score to 1 if it should be included
        classification_score = tf.cond(tf.cast(include_classification_score, tf.bool),
                                        true_fn=lambda: classification_score,
                                        false_fn=lambda: tf.ones_like(classification_score))

        # Compute score as defined in DRise for all possible pair of boxes
        # (nb_box_ref, nb_box_pred)
        boxes_pairwise_scores = intersection_score \
                                * detection_probability \
                                * classification_score

        # select for a reference box the most similar predicted box score
        # (nb_box_ref,)
        ref_boxes_scores = tf.reduce_max(boxes_pairwise_scores, axis=1)

        # get an attribution for several boxes in the same time
        # ()
        image_score = tf.reduce_mean(ref_boxes_scores)
        return image_score

    objects = model(inputs)
    return tf.map_fn(batch_loop, (objects, targets), fn_output_signature=tf.float32)



"""
Pretty plots option for explanations
"""
from math import ceil

import tensorflow as tf
import numpy as np
from matplotlib import pyplot as plt

from ..types import Optional, Union


def _normalize(image: Union[tf.Tensor, np.ndarray]) -> np.ndarray:
    """
    Normalize an image in [0, 1].

    Parameters
    ----------
    image
        Image to prepare.

    Returns
    -------
    image
        Image ready to be used with matplotlib (in range[0, 1]).
    """
    image = np.array(image, np.float32)

    image -= image.min()
    image /= image.max()

    return image

def _clip_percentile(tensor: Union[tf.Tensor, np.ndarray],
                     percentile: float) -> np.ndarray:
    """
    Apply clip according to percentile value (percentile, 100-percentile) of a tensor
    only if percentile is not None.

    Parameters
    ----------
    tensor
        tensor to clip.

    Returns
    -------
    tensor_clipped
        Tensor clipped accordingly to the percentile value.
    """

    assert 0. <= percentile <= 100., "Percentile value should be in [0, 100]"

    if percentile is not None:
        clip_min = np.percentile(tensor, percentile)
        clip_max = np.percentile(tensor, 100. - percentile)
        tensor = np.clip(tensor, clip_min, clip_max)

    return tensor


def plot_attribution(explanation,
                      image: Optional[np.ndarray] = None,
                      cmap: str = "jet",
                      alpha: float = 0.5,
                      clip_percentile: Optional[float] = 0.1,
                      absolute_value: bool = False,
                      **plot_kwargs):
    """
    Displays a single explanation and the associated image (if provided).
    Applies a series of pre-processing to facilitate the interpretation of heatmaps.

    Parameters
    ----------
    explanation
        Attribution / heatmap to plot.
    image
        Image associated to the explanations.
    cmap
        Matplotlib color map to apply.
    alpha
        Opacity value for the explanation.
    clip_percentile
        Percentile value to use if clipping is needed, e.g a value of 1 will perform a clipping
        between percentile 1 and 99. This parameter allows to avoid outliers  in case of too
        extreme values.
    absolute_value
        Whether an absolute value is applied to the explanations.
    plot_kwargs
        Additional parameters passed to `plt.imshow()`.
    """
    if image is not None:
        image = _normalize(image)
        plt.imshow(image)

    if explanation.shape[-1] == 3:
        explanation = np.mean(explanation, -1)

    if absolute_value:
        explanation = np.abs(explanation)

    if clip_percentile:
        explanation = _clip_percentile(explanation, clip_percentile)

    explanation = _normalize(explanation)

    plt.imshow(explanation, cmap=cmap, alpha=alpha, **plot_kwargs)
    plt.axis('off')


def plot_attributions(
        explanations: Union[tf.Tensor, np.ndarray],
        images: Optional[Union[tf.Tensor, np.ndarray]] = None,
        cmap: str = "viridis",
        alpha: float = 0.5,
        clip_percentile: Optional[float] = 0.1,
        absolute_value: bool = False,
        cols: int = 5,
        img_size: float = 2.,
        **plot_kwargs
):
    """
    Displays a series of explanations and their associated images if these are provided.
    Applies pre-processing to facilitate the interpretation of heatmaps.

    Parameters
    ----------
    explanations
        Attributions values to plot.
    images
        Images associated to explanations. If provided, there must be one explanation for each
        image.
    cmap
        Matplotlib color map to apply.
    alpha
        Opacity value for the explanation.
    clip_percentile
        Percentile value to use if clipping is needed, e.g a value of 1 will perform a clipping
        between percentile 1 and 99. This parameter allows to avoid outliers  in case of too
        extreme values.
    absolute_value
        Whether an absolute value is applied to the explanations.
    cols
        Number of columns.
    img_size:
        Size of each subplots (in inch), considering we keep aspect ratio
    plot_kwargs
        Additional parameters passed to `plt.imshow()`.
    """
    if images is not None:
        assert len(images) == len(explanations), "If you provide images, there must be as many " \
                                                 "as explanations."

    rows = ceil(len(explanations) / cols)
    # get width and height of our images
    l_width, l_height = explanations.shape[1:3]

    # define the figure margin, width, height in inch
    margin = 0.3
    spacing = 0.3
    figwidth = cols * img_size + (cols-1) * spacing + 2 * margin
    figheight = rows * img_size * l_height/l_width + (rows-1) * spacing + 2 * margin

    left = margin/figwidth
    bottom = margin/figheight

    fig = plt.figure()
    fig.set_size_inches(figwidth, figheight)

    fig.subplots_adjust(
        left = left,
        bottom = bottom,
        right = 1.-left,
        top = 1.-bottom,
        wspace = spacing/img_size,
        hspace= spacing/img_size * l_width/l_height
    )

    for i, explanation in enumerate(explanations):
        plt.subplot(rows, cols, i+1)

        if images is not None:
            img = _normalize(images[i])
            if img.shape[-1] == 1:
                plt.imshow(img[:,:,0], cmap="Greys")
            else:
                plt.imshow(img)

        plot_attribution(explanation, cmap=cmap, alpha=alpha, clip_percentile=clip_percentile,
                         absolute_value=absolute_value, **plot_kwargs)

def plot_maco(image, alpha, percentile_image=1.0, percentile_alpha=80):
    """
    Plot maco feature visualization image (take care of merging the alpha).

    Parameters
    ----------
    image
        Image to plot.
    alpha
        Alpha channel to plot.
    percentile_image
        Percentile value to use to ceil the image and avoid extreme values.
    percentile_alpha
        Percentile value to use to ceil the alpha channel. A higher value will result in a more
        transparent image with only the most important features.
    """

    image = np.array(image).copy()
    image = _clip_percentile(image, percentile_image)

    alpha = np.mean(np.array(alpha).copy(), -1, keepdims=True)
    alpha = np.clip(alpha, 0, np.percentile(alpha, percentile_alpha))
    alpha = alpha / alpha.max()

    image = image * alpha
    image = _normalize(image)

    plt.imshow(np.concatenate([image, alpha], -1))
    plt.axis('off')



"""
Test object detection BoundingBoxesExplainer
"""
import os
import sys
sys.path.append(os.getcwd())

import unittest

import tensorflow as tf

from tests.utils import generate_data, generate_object_detection_model

from xplique.commons.exceptions import InvalidModelException
from xplique.attributions import BoundingBoxesExplainer, Rise, Saliency


def test_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    method = Rise(model, nb_samples=10)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    explainer = BoundingBoxesExplainer(method)

    test_raise_assertion_error = unittest.TestCase().assertRaises
    test_raise_assertion_error(InvalidModelException, explainer.gradient)
    test_raise_assertion_error(InvalidModelException, explainer.batch_gradient)

    phis = explainer(x, obj_ref)

    assert phis.shape == (obj_ref.shape[0], input_shape[0], input_shape[1])


def test_gradient_object_detector():
    """Assert input shape returned is correct"""
    input_shape = (8, 8, 1)
    nb_samples = 3
    nb_labels = 2
    max_nb_boxes = 4
    x, _ = generate_data(input_shape, nb_labels, nb_samples)
    model = generate_object_detection_model(input_shape, max_nb_boxes=max_nb_boxes, nb_labels=nb_labels)

    method = Saliency(model)

    obj_ref = tf.cast([
                [0, 0, 100, 100, 0.9, 1.0, 0.0],
                [50, 50, 150, 150, 0.5, 1.0, 0.0],
                [0, 10, 20, 30, 0.7, 0.0, 1.0],
            ], tf.float32)

    explainer = BoundingBoxesExplainer(method)

    phis = explainer(x, obj_ref)

    assert phis.shape == (obj_ref.shape[0], input_shape[0], input_shape[1])


import numpy as np

from xplique.attributions import Occlusion
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output shape must be the same as the input shape, except for the channels"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 20)
        model = generate_model(input_shape, nb_labels)

        method = Occlusion(model)
        sensitivity = method.explain(x, y)

        assert x.shape[:3] == sensitivity.shape


def test_polymorphic_parameters():
    """Ensure we could pass tuple or int to define patch parameters when inputs are images"""
    s = 3

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        features, targets = generate_data(input_shape, nb_labels, 20)
        model = generate_model(input_shape, nb_labels)

        occlusion_int = Occlusion(model, patch_size=s, patch_stride=s)
        occlusion_tuple = Occlusion(model, patch_size=(s, s), patch_stride=(s, s))

        occlusion_int(features, targets)
        occlusion_tuple(features, targets)

        assert occlusion_int.patch_size == occlusion_tuple.patch_size
        assert occlusion_int.patch_stride == occlusion_tuple.patch_stride


def test_mask_generator():
    """Ensure we generate all the possible masks"""

    assert Occlusion._get_masks((10, 10), (1, 1), (1, 1)).shape == (100, 10, 10)
    assert Occlusion._get_masks((10, 10), (2, 2), (2, 2)).shape == (25, 10, 10)
    assert Occlusion._get_masks((10, 10), (2, 2), (3, 3)).shape == (9, 10, 10)
    assert np.array_equal(Occlusion._get_masks((2, 2), (1, 1), (1, 1)), np.array([
        [
            [1, 0],
            [0, 0]
        ],
        [
            [0, 1],
            [0, 0]
        ],
        [
            [0, 0],
            [1, 0]
        ],
        [
            [0, 0],
            [0, 1]
        ],
    ], dtype=np.float32))


def test_apply():
    """Ensure we apply correctly the masks"""
    x = np.ones((2, 2, 1), dtype=np.float32)
    applied_value = 49.0

    masks = Occlusion._get_masks((2, 2, 1), (1, 1), (1, 1))
    occluded_x = Occlusion._apply_masks(x, masks, applied_value)

    assert almost_equal(occluded_x[:,:,:,0], np.array([
        [[applied_value, 1.0],
         [1.0, 1.0]],
        [[1.0, applied_value],
         [1.0, 1.0]],
        [[1.0, 1.0],
         [applied_value, 1.0]],
        [[1.0, 1.0],
         [1.0, applied_value]],
    ]))

    masks = Occlusion._get_masks((2, 2, 1), (2, 2), (2, 2))
    occluded_x = Occlusion._apply_masks(x, masks, applied_value)

    assert almost_equal(occluded_x[:, :, :, 0], np.array([
        [[applied_value, applied_value],
         [applied_value, applied_value]]
    ]))


def test_delta_computation():
    """Ensure get correct delta score"""
    baseline_score = np.array([1.0], dtype=np.float32)
    masks = Occlusion._get_masks((2, 2, 1), (1, 1), (1, 1))
    scores = np.array([1.0, 0.0, 10.0, -1.0], dtype=np.float32)

    deltas = Occlusion._compute_sensitivity(baseline_score, scores, masks)
    assert almost_equal(deltas.numpy().flatten(), [0.0, 1.0, -9.0, 2.0])

    baseline_score = np.array([10.0], dtype=np.float32)
    masks = Occlusion._get_masks((2, 2, 1), (2, 2), (2, 2))
    scores = np.array([5.0], dtype=np.float32)

    deltas = Occlusion._compute_sensitivity(baseline_score, scores, masks)
    assert almost_equal(deltas, [5.0])



"""
Module for having a wrapper for PyTorch models
"""
import warnings

import tensorflow as tf
import numpy as np

from ..types import Union, Optional, Tuple, Callable

class TorchWrapper(tf.keras.Model):
    """
    A wrapper for PyTorch models so that they can be used in Xplique framework
    for most attribution methods

    Parameters
    ----------
    torch_model
        A nn.Module PyTorch model
    device
        If we are on GPU or CPU
    is_channel_first
        A boolean that is true if the torch's model expect a channel dim and if this one come first
    """

    def __init__(self, torch_model: "nn.Module", device: Union["torch.device", str],
                 is_channel_first: Optional[bool] = None
                 ): # pylint: disable=C0415,C0103

        super().__init__()

        try:
            # use PyTorch functionality
            import torch
            from torch import nn
            from torch import from_numpy
            self.torch = torch
            self.nn = nn
            self.from_numpy = from_numpy
        except ImportError as exc:
            raise ImportError(
            "PyTorch is required to use this feature. \
             Please install PyTorch using 'pip install torch'."
            ) from exc

        assert not(torch_model.training), "Please provide a torch module in eval mode"
        self.model = torch_model.to(device)
        self.device = device
        # with torch, the convention for CNN is (N, C, H, W)
        if is_channel_first is None:
            self.channel_first = self._has_conv_layers()
        else:
            self.channel_first = is_channel_first
        # deactivate all tf.function
        tf.config.run_functions_eagerly(True)
        warnings.warn("TF is set to run eagerly to avoid conflict with PyTorch. Thus,\
                       TF functions might be slower")

    # pylint: disable=arguments-differ
    @tf.custom_gradient
    def call(self, inputs: np.ndarray) -> Tuple[tf.Tensor, Callable]:
        """
        A method that allow to call the PyTorch model wrapped on inputs that are
        numpy arrays reshaped to match Xplique's explainers expectations (i.e. channel-last)
        , get the gradients from PyTorch framework and recast outputs and gradients
        as Tensorflow tensors.

        Parameters
        ----------
        inputs
            Processed inputs as numpy arrays

        Returns
        -------
        outputs
            Outputs of the PyTorch model converted to Tensorflow tensor
        grad
            The function that allow to compute the gradient of the PyTorch model and
            broadcast it for Tensorflow
        """
        # transform your numpy inputs to torch
        torch_inputs = self.np_img_to_torch(inputs).to(self.device)
        torch_inputs.requires_grad_(True)

        # make predictions
        self.model.zero_grad()
        outputs = self.model(torch_inputs)
        output_tensor = tf.constant(outputs.cpu().detach().numpy())

        def grad(upstream):
            self.torch.autograd.backward(
                outputs,
                grad_tensors=self.from_numpy(upstream.numpy()).to(self.device),
                retain_graph=False
            )
            dx_torch = torch_inputs.grad

            dx_np = dx_torch.cpu().detach().numpy()
            if self.channel_first:
                # (N, C, H, W) -> (N, H, W, C) for explainer
                dx_np = np.moveaxis(dx_np, [1, 2, 3], [3, 1, 2])

            gradient = tf.constant(dx_np)
            return gradient

        return output_tensor, grad

    def np_img_to_torch(self, np_inputs: np.ndarray):
        """
        Methods that transform inputs as expected by the explainer to inputs expected
        by your PyTorch model.

        Notes
        -----
        If your model (or your data pipeline) has more specifities, change this function to make
        the bridge

        Parameters
        ----------
        np_inputs
            Inputs as numpy arrays (simpler for conversion between PyTorch
            and Tensorflow)

        Returns
        -------
        torch_inputs
            Torch Tensor with the format expected by the PyTorch model
        """
        if self.channel_first:
            # go from channel last to channel first for the torch model
            np_inputs = np.moveaxis(np_inputs, [3, 1, 2], [1, 2, 3])
        else:
            np_inputs = np.asarray(np_inputs)

        # convert numpy array to torch tensor
        torch_inputs = self.torch.Tensor(np_inputs)

        return torch_inputs

    def _has_conv_layers(self):
        """
        A method that checks if the PyTorch models has 2D convolutional layer.
        Indeed, convolution with PyTorch expects inputs in the shape (N, C, H, W)
        where TF expect (N, H, W, C).

        Returns
        -------
        has_conv_layers
            A boolean that says if the PyTorch model has Conv2d layers
        """
        has_conv_layers = False

        for module in self.model.modules():
            if isinstance(module, self.nn.Conv2d):
                has_conv_layers = True
                break

        return has_conv_layers



"""
Objective wrapper and utils to build a function to be optimized
"""

import itertools

import tensorflow as tf
import numpy as np

from ..commons import find_layer
from ..types import Union, List, Callable, Tuple, Optional
from .losses import dot_cossim


class Objective:
    """
    Use to combine several sub-objectives into one.

    Each sub-objective act on a layer, possibly on a neuron or a channel (in
    that case we apply a mask on the layer values), or even multiple neurons (in
    that case we have multiples masks). When two sub-objectives are added, we
    optimize all their combinations.

    e.g Objective 1 target the neurons 1 to 10 of the logits l1,...,l10
        Objective 2 target a direction on the first layer d1
        Objective 3 target each of the 5 channels on another layer c1,...,c5

        The resulting Objective will have 10*5*1 combinations. The first input
        will optimize l1+d1+c1 and the last one l10+d1+c5.

    Parameters
    ----------
    model
        Model used for optimization.
    layers
        A list of the layers output for each sub-objectives.
    masks
        A list of masks that will be applied on the targeted layer for each
        sub-objectives.
    funcs
        A list of loss functions for each sub-objectives.
    multipliers
        A list of multiplication factor for each sub-objectives
    names
        A list of name for each sub-objectives
    """

    def __init__(self,
                 model: tf.keras.Model,
                 layers: List[tf.keras.layers.Layer],
                 masks: List[tf.Tensor],
                 funcs: List[Callable],
                 multipliers: List[float],
                 names: List[str]):
        self.model = model
        self.layers = layers
        self.masks = masks
        self.funcs = funcs
        self.multipliers = multipliers
        self.names = names

    def __add__(self, term):
        if not isinstance(term, Objective):
            raise ValueError(f"{term} is not an objective.")
        return Objective(
                self.model,
                layers=self.layers + term.layers,
                masks=self.masks + term.masks,
                funcs=self.funcs + term.funcs,
                multipliers=self.multipliers + term.multipliers,
                names=self.names + term.names
        )

    def __sub__(self, term):
        if not isinstance(term, Objective):
            raise ValueError(f"{term} is not an objective.")
        term.multipliers = [-1.0 * m for m in term.multipliers]
        return self + term

    def __mul__(self, factor: float):
        if not isinstance(factor, (int, float)):
            raise ValueError(f"{factor} is not a number.")
        self.multipliers = [m * factor for m in self.multipliers]
        return self

    def __rmul__(self, factor: float):
        return self * factor

    def compile(self) -> Tuple[tf.keras.Model, Callable, List[str], Tuple]:
        """
        Compile all the sub-objectives into one and return the objects
        for the optimisation process.

        Returns
        -------
        model_reconfigured
            Model with the outputs needed for the optimization.
        objective_function
            Function to call that compute the loss for the objectives.
        names
            Names of each objectives.
        input_shape
            Shape of the input, one sample for each optimization.
        """
        # the number of inputs will be the number of combinations possible
        # of the objectives, the mask are used to take into account
        # these combinations
        nb_sub_objectives = len(self.multipliers)

        # re-arrange to match the different objectives with the model outputs
        masks = np.array([np.array(m, dtype=object) for m in itertools.product(*self.masks)])
        masks = [tf.cast(tf.stack(list(masks[:, i])), tf.float32) for i in
                 range(nb_sub_objectives)]

        # the name of each combination is the concatenation of each objectives
        names = np.array([' & '.join(names) for names in
                           itertools.product(*self.names)])
        # one multiplier by sub-objective
        multipliers = tf.constant(self.multipliers)

        def objective_function(model_outputs):
            loss = 0.0
            for output_index in range(0, nb_sub_objectives):
                loss += self.funcs[output_index](model_outputs[output_index],
                                                 masks[output_index]) * \
                                                 multipliers[output_index]
            return loss

        # the model outputs will be composed of the layers needed
        model_reconfigured = tf.keras.Model(self.model.input, [*self.layers])

        nb_combinations = masks[0].shape[0]
        input_shape = (nb_combinations, *model_reconfigured.input.shape[1:])

        return model_reconfigured, objective_function, names, input_shape

    @staticmethod
    def layer(model: tf.keras.Model,
              layer: Union[str, int],
              reducer: str = "magnitude",
              multiplier: float = 1.0,
              name: Optional[str] = None):
        """
        Util to build an objective to maximise a layer.

        Parameters
        ----------
        model
            Model used for optimization.
        layer
            Index or name of the targeted layer.
        reducer
            Type of reduction to apply, 'mean' will optimize the mean value of the
            layer, 'magnitude' will optimize the mean of the absolute values.
        multiplier
            Multiplication factor of the objective.
        name
            A name for the objective.

        Returns
        -------
        objective
            An objective ready to be compiled
        """
        layer = find_layer(model, layer)
        layer_shape = layer.output.shape
        mask = np.ones((1, *layer_shape[1:]))

        if name is None:
            name = [f"Layer#{layer.name}"]

        power = 2.0 if reducer == "magnitude" else 1.0

        def optim_func(model_output, mask):
            return tf.reduce_mean((model_output * mask) ** power)

        return Objective(model, [layer.output], [mask], [optim_func], [multiplier], [name])

    @staticmethod
    def direction(model: tf.keras.Model,
                  layer: Union[str, int],
                  vectors: Union[tf.Tensor, List[tf.Tensor]],
                  multiplier: float = 1.0,
                  cossim_pow: float = 2.0,
                  names: Optional[Union[str, List[str]]] = None):
        """
        Util to build an objective to maximise a direction of a layer.

        Parameters
        ----------
        model
            Model used for optimization.
        layer
            Index or name of the targeted layer.
        vectors
            Direction(s) to optimize.
        multiplier
            Multiplication factor of the objective.
        cossim_pow
            Power of the cosine similarity, higher value encourage the objective to care more about
            the angle of the activations.
        names
            A name for each objectives.

        Returns
        -------
        objective
            An objective ready to be compiled
        """
        layer = find_layer(model, layer)
        masks = vectors if isinstance(vectors, list) else [vectors]

        if names is None:
            names = [f"Direction#{layer.name}_{i}" for i in range(len(masks))]

        def optim_func(model_output, mask):
            return dot_cossim(model_output, mask, cossim_pow)

        return Objective(model, [layer.output], [masks], [optim_func], [multiplier], [names])

    @staticmethod
    def channel(model: tf.keras.Model,
                layer: Union[str, int],
                channel_ids: Union[int, List[int]],
                multiplier: float = 1.0,
                names: Optional[Union[str, List[str]]] = None):
        """
        Util to build an objective to maximise a channel.

        Parameters
        ----------
        model
            Model used for optimization.
        layer
            Index or name of the targeted layer.
        channel_ids
            Indexes of the channels to maximise.
        multiplier
            Multiplication factor of the objectives.
        names
            Names for each objectives.

        Returns
        -------
        objective
            An objective containing a sub-objective for each channels.
        """
        layer = find_layer(model, layer)
        layer_shape = layer.output.shape
        channel_ids = channel_ids if isinstance(channel_ids, list) else [channel_ids]

        # for each targeted channel, create a boolean mask on the layer to target
        # the channel
        masks = np.zeros((len(channel_ids), *layer_shape[1:]))
        for i, c_id in enumerate(channel_ids):
            masks[i, ..., c_id] = 1.0

        if names is None:
            names = [f"Channel#{layer.name}_{ch_id}" for ch_id in channel_ids]

        axis_to_reduce = list(range(1, len(layer_shape)))

        def optim_func(output, target):
            return tf.reduce_mean(output * target, axis=axis_to_reduce)

        return Objective(model, [layer.output], [masks], [optim_func], [multiplier], [names])

    @staticmethod
    def neuron(model: tf.keras.Model,
               layer: Union[str, int],
               neurons_ids: Union[int, List[int]],
               multiplier: float = 1.0,
               names: Optional[Union[str, List[str]]] = None):
        """
        Util to build an objective to maximise a neuron.

        Parameters
        ----------
        model
            Model used for optimization.
        layer
            Index or name of the targeted layer.
        neurons_ids
            Indexes of the neurons to maximise.
        multiplier
            Multiplication factor of the objectives.
        names
            Names for each objectives.

        Returns
        -------
        objective
            An objective containing a sub-objective for each neurons.
        """
        layer = find_layer(model, layer)

        neurons_ids = neurons_ids if isinstance(neurons_ids, list) else [neurons_ids]
        nb_objectives = len(neurons_ids)
        layer_shape = layer.output.shape[1:]

        # for each targeted neurons, create a boolean mask on the layer to target it
        masks = np.zeros((nb_objectives, *layer_shape))
        masks = masks.reshape((nb_objectives, -1))
        for i, neuron_id in enumerate(neurons_ids):
            masks[i, neuron_id] = 1.0
        masks = masks.reshape((nb_objectives, *layer_shape))

        if names is None:
            names = [f"Neuron#{layer.name}_{neuron_id}" for neuron_id in neurons_ids]

        axis_to_reduce = list(range(1, len(layer_shape)+1))

        def optim_func(output, target):
            return tf.reduce_mean(output * target, axis=axis_to_reduce)

        return Objective(model, [layer.output], [masks], [optim_func], [multiplier], [names])



from xplique.attributions import GradientInput
from ..utils import generate_data, generate_model


def test_output_shape():
    """The output shape must be the same as the input shape"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 100)
        model = generate_model(input_shape, nb_labels)

        method = GradientInput(model, -2)
        outputs = method.explain(x, y)

        assert x.shape == outputs.shape



"""
Test utils functions for object detection
"""

import numpy as np

from xplique.utils_functions.object_detection import _box_iou

from ..utils import almost_equal


def test_iou_mask():
    """Assert the Mask IoU calculation is ok"""
    dtype = np.float32

    m1 = np.array([
        [10, 20, 30, 40]
    ], dtype=dtype)
    m2 = np.array([
        [15, 20, 30, 40]
    ], dtype=dtype)
    m3 = np.array([
        [0, 20, 10, 40]
    ], dtype=dtype)
    m4 = np.array([
        [0, 0, 100, 100]
    ], dtype=dtype)

    assert almost_equal(_box_iou(m1, m2), 300.0 / 400.0)
    assert almost_equal(_box_iou(m1, m3), 0.0)
    assert almost_equal(_box_iou(m3, m2), 0.0)
    assert almost_equal(_box_iou(m1, m4), 400.0 / 10_000)
    assert almost_equal(_box_iou(m2, m4), 300.0 / 10_000)
    assert almost_equal(_box_iou(m3, m4), 200.0 / 10_000)



import numpy as np
import tensorflow as tf

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, GradCAM, Occlusion, Rise, GuidedBackprop, DeconvNet,
                                  GradCAMPP, Lime, KernelShap, SobolAttributionMethod,
                                  HsicAttributionMethod)
from xplique.attributions.base import BlackBoxExplainer
from ..utils import generate_data, generate_model


def _default_methods(model, output_layer_index):
    return [
        Saliency(model, output_layer_index),
        GradientInput(model, output_layer_index),
        SmoothGrad(model, output_layer_index),
        VarGrad(model, output_layer_index),
        SquareGrad(model, output_layer_index),
        IntegratedGradients(model, output_layer_index),
        GradCAM(model, output_layer_index),
        Occlusion(model),
        Rise(model, nb_samples=2),
        GuidedBackprop(model, output_layer_index),
        DeconvNet(model, output_layer_index),
        GradCAMPP(model, output_layer_index),
        Lime(model, nb_samples=2),
        KernelShap(model, nb_samples=2),
        SobolAttributionMethod(model, grid_size=2, nb_design=2),
        HsicAttributionMethod(model, grid_size=2, nb_design=2),
    ]


def test_common():
    """Test applied to all the attributions"""

    input_shape, nb_labels, samples = ((32, 32, 3), 10, 20)
    model = generate_model(input_shape, nb_labels)
    output_layer_index = -2

    inputs_np, targets_np = generate_data(input_shape, nb_labels, samples)
    inputs_tf, targets_tf = tf.cast(inputs_np, tf.float32), tf.cast(targets_np, tf.float32)
    dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np))
    batched_dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np)).batch(3)

    methods = _default_methods(model, output_layer_index)

    for inputs, targets in [(inputs_np, targets_np),
                            (inputs_tf, targets_tf),
                            (dataset, None),
                            (batched_dataset, None)]:
        for method in methods:
            explanations = method.explain(inputs, targets)

            # all explanation must have an explain method
            assert hasattr(method, 'explain')

            # all explanations returned must be either a tf.Tensor or ndarray
            assert isinstance(explanations, (tf.Tensor, np.ndarray))

            # we should have one explanation for each inputs
            assert len(explanations) == len(inputs_np)


def test_batch_size():
    """Ensure the functioning of attributions for special batch size cases"""

    input_shape, nb_labels, samples = ((10, 10, 3), 5, 20)
    inputs, targets = generate_data(input_shape, nb_labels, samples)
    model = generate_model(input_shape, nb_labels)
    output_layer_index = -1

    batch_sizes = [None, 1, 32]

    for bs in batch_sizes:

        methods = [
            Saliency(model, output_layer_index, bs),
            GradientInput(model, output_layer_index, bs),
            SmoothGrad(model, output_layer_index, bs),
            VarGrad(model, output_layer_index, bs),
            SquareGrad(model, output_layer_index, bs),
            IntegratedGradients(model, output_layer_index, bs),
            GradCAM(model, output_layer_index, bs),
            Occlusion(model, bs),
            Rise(model, bs, nb_samples=2),
            GuidedBackprop(model, output_layer_index, bs),
            DeconvNet(model, output_layer_index, bs),
            GradCAMPP(model, output_layer_index, bs),
            Lime(model, bs, nb_samples=2),
            KernelShap(model, bs, nb_samples=2),
            SobolAttributionMethod(model, grid_size=2, nb_design=2)
        ]

        for method in methods:
            explanations = method.explain(inputs, targets)

            # all explanations returned must be either a tf.Tensor or ndarray
            assert isinstance(explanations, (tf.Tensor, np.ndarray))

            # we should have one explanation for each inputs
            assert len(explanations) == len(inputs)


def test_model_caching():
    """Test the caching engine, used to avoid re-tracing"""

    model = generate_model()
    output_layer_index = -1

    # the key used for caching is the following tuple
    cache_key = (id(model.input), id(model.output))

    cache_len_before = len(BlackBoxExplainer._cache_models.keys())  # pylint:
    # disable=protected-access

    assert (cache_key not in BlackBoxExplainer._cache_models)  # pylint: disable=protected-access

    _ = _default_methods(model, output_layer_index)

    # check that the key is now in the cache
    assert (cache_key in BlackBoxExplainer._cache_models)  # pylint: disable=protected-access

    # ensure that there no more than one key has been added
    assert (len(
        BlackBoxExplainer._cache_models) == cache_len_before + 1)  # pylint: disable=protected-access


"""
Plots for tabular data
"""
import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np

def _sanitize_features_name(explanations, features_name):
    """
    Helper that provide generic features name (with the feature index) if features name
    is None.
    """
    if features_name is None:
        single_explanation = len(explanations.shape)==1
        if single_explanation:
            features_name = [f"Feature {str(j)}" for j in range(len(explanations))]
        else:
            features_name = [f"Feature {str(j)}" for j in range(explanations.shape[1])]
    return features_name

def _select_features(explanations, max_display, features_name):
    """
    Helper, that select max_display features. Useful if the number of features is huge.
    We keep the features which have a mean absolute impact greater than the others.
    """
    # if we have a lot of feature we will keep only max display
    if max_display is None:
        num_features_kept = len(features_name)
        features_idx_kept = np.arange(num_features_kept)
    else:
        num_features_kept = min(max_display, len(features_name))
        single_explanation = len(explanations.shape)==1
        if single_explanation:
            ranked = np.argsort(
               np.abs(explanations)
            )
        else:
            ranked = np.argsort(
                np.mean(np.abs(explanations), axis=0)
            )
        features_idx_kept = ranked[::-1][:num_features_kept]
    return num_features_kept, features_idx_kept

def _add_colorbar(cmap):
    """
    Add the color bar corresponding to cmap to the plot
    """
    mappa = cm.ScalarMappable(cmap=cmap)
    colorbar = plt.colorbar(mappa, ticks=[0, 1])
    colorbar.set_ticklabels(['Low', 'High'])
    colorbar.set_label('Feature value', size=12, labelpad=0)
    colorbar.ax.tick_params(labelsize=11, length=0)
    colorbar.set_alpha(1)
    colorbar.outline.set_visible(False)
    bbox = colorbar.ax.get_window_extent().transformed(plt.gcf().dpi_scale_trans.inverted())
    colorbar.ax.set_aspect((bbox.height - 0.9) * 20)

def _clip_values(fvalues, clip_percentile):
    """
    Clip value of fvalues between clip_percentile and 100-clip_percentile
    """
    if clip_percentile is not None:
        clip_min = np.percentile(fvalues, clip_percentile)
        clip_max = np.percentile(fvalues, 100 - clip_percentile)
        fvalues = np.clip(fvalues, clip_min, clip_max)
    return fvalues

def _get_offset_positions(explanation_val_feature, nb_points, row_height):
    """
    Helper to set an offset position on the y-axis for points with close x-value.


    Paramerers
    ----------
    explanation_val_feature
        All the explanation value for a single feature
    nb_points
        The number of explanation value per feature
    row_height
        The length of the available y-axis in which points will be displayed

    Returns
    -------
    offset
        The y position offset for each points
    """
    nbins = 100 # nb of division on the y-axis for one feature

    # get the points where explanation values are close
    max_feat_val = np.max(explanation_val_feature)
    min_feat_val = np.min(explanation_val_feature)
    num = nbins * (explanation_val_feature - min_feat_val)
    den = max_feat_val - min_feat_val + 1e-8 # in case max=min
    # points with the same quant value might overlap
    quant = np.round(num/den)

    # here we make sure points with the same quant value have different offset
    inds = np.argsort(quant + np.random.randn(nb_points) * 1e-6)
    layer = 0
    last_bin = -1
    offset = np.zeros(nb_points)
    for ind in inds:
        # check if the two points are close and might overlap
        if quant[ind] != last_bin:
            layer = 0
        # centered around the line, one time over the other below
        offset[ind] = np.ceil(layer / 2) * ((layer % 2) * 2 - 1)
        layer += 1
        last_bin = quant[ind]
    # adjust the offsets to the row height
    offset *= 0.9 * (row_height / np.max(offset + 1))

    return offset

def plot_feature_impact(
    explanation,
    features_name=None,
    features_value=None,
    max_display=None,
):
    """
    Bar plot of the explanation values. Some options are directly inspired from the
    [shap library](https://github.com/slundberg/shap).

    Parameters
    ----------
    explanation
        A single explanation
    features_name
        The name of the features. If not provided will be displayed as: Feature: idx_feature
    features_value
        The value of features from which the explanation was computed. If provided, it will be
        displayed next to the corresponding label
    max_display
        If there is too many features, set this parameter to display only the max_display features
        with the most significant impact on the output
    """
    # sanitize to numpy array
    explanation = np.array(explanation)

    # add default features name if not provided
    features_name = _sanitize_features_name(explanation, features_name)

    # if we have a lot of feature we will keep only max display
    num_features_kept, features_idx_kept = _select_features(explanation, max_display, features_name)
    explanation_kept = explanation[features_idx_kept]

    # build y-ticks label
    yticklabels = []
    for idx_kept in features_idx_kept:
        if features_value is None:
            yticklabels.append(features_name[idx_kept])
        else:
            yticklabels.append(
                f"{str(round(features_value[idx_kept], 3))} = {features_name[idx_kept]}"
            )

    y_pos = np.arange(num_features_kept)  # the label locations

    fig, axes = plt.subplots()

    # the actual bar plot
    colors = [
        'slateblue' if explanation_kept[j]<=0 else 'yellowgreen' for j in range(num_features_kept)
    ]
    axes.barh(
        y_pos,
        explanation_kept,
        align='center',
        color=colors
    )

    # add the explanation value next to the bar
    xlen = plt.xlim()[1] - plt.xlim()[0]
    bbox = axes.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    xscale = xlen/bbox.width

    for i in y_pos:
        # put horizontal lines for each feature row
        plt.axhline(i, color="darkgrey", lw=0.5, dashes=(1, 5), zorder=-1)
        if explanation_kept[i] < 0:
            axes.text(
                explanation_kept[i] - 0.02*xscale,
                y_pos[i],
                f"{str(round(explanation_kept[i], 2))}",
                horizontalalignment='right',
                verticalalignment='center',
                fontsize=10
            )
        else:
            axes.text(
                explanation_kept[i] + 0.02*xscale,
                y_pos[i],
                f"{str(round(explanation_kept[i], 2))}",
                horizontalalignment='left',
                verticalalignment='center',
                fontsize=10
            )

    # add some text for labels and custom y-axis tick labels
    axes.set_xlabel('Impact on output')
    axes.set_ylabel('')
    axes.set_title('Features impact')
    axes.set_yticks(y_pos)
    axes.set_yticklabels(yticklabels)

    # make the plot prettier
    fig.tight_layout()

    axes.spines['right'].set_visible(False)
    axes.spines['top'].set_visible(False)
    xmin,xmax = plt.gca().get_xlim()

    # if we have negative values draw a vertical axis at zero
    neg_val = len(explanation_kept[explanation_kept<0])!=0
    if neg_val:
        plt.axvline(0, 0, 1, color="dimgray", linestyle="-", linewidth=1)
        axes.spines['left'].set_visible(False)
        plt.gca().set_xlim(xmin - (xmax-xmin)*0.1, xmax + (xmax-xmin)*0.1)
    else:
        plt.gca().set_xlim(xmin, xmax + (xmax-xmin)*0.1)

def plot_mean_feature_impact(
    explanations,
    features_name=None,
    max_display=None,
):
    """
    The same than plot_feature_importance but we will consider the mean explanation value
    grouped by feature.
    A more informative plot is the summary_plot_tabular.
    """
    # sanitize explanations to numpy array
    explanations = np.array(explanations)

    mean_explanation_per_feature = np.mean(explanations, axis=0)

    plot_feature_impact(
        mean_explanation_per_feature,
        features_name=features_name,
        max_display=max_display
    )

def summary_plot_tabular(
    explanations,
    features_values = None,
    features_name = None,
    max_display = None,
    cmap = 'viridis',
    clip_percentile = 1,
    alpha = 1,
    plot_size = None,
):
    """
    Summary plot adapted from the [shap library](https://github.com/slundberg/shap). Usefull to
    have an overall idea of the impact of each features on the output depending on their value.

    Parameters
    ----------
    explanations
        The explanations from which we want to draw insights
    features_values
        The features values of each samples leading to the different explanations.
        Shape must match explanations shape.
    features_name
        The name of the features. If not provided will be displayed as: Feature: idx_feature
    max_display
        If there is too many features, set this parameter to display only the max_display features
        with the most significant impact on the output
    cmap
        Matplotlib color map to apply.
    clip_percentile
        Percentile value to use if clipping is needed, e.g a value of 1 will perform a clipping
        between percentile 1 and 99. This parameter allows to avoid outliers  in case of too
        extreme values.
    alpha
        Opacity value for the feature value.
    plot_size
        If provided as a tuple or a list will set the figure size in inches.
        If provided as an int, will define the row_height for each features.
    """
    # sanitize to numpy array
    explanations = np.array(explanations)

    # add default features name if not provided
    features_name = _sanitize_features_name(explanations, features_name)

    # if we have a lot of feature we will keep only max display
    nb_features_kept, features_idx_kept = _select_features(explanations, max_display, features_name)

    # build our y-tick labels
    yticklabels = [features_name[idx_kept] for idx_kept in features_idx_kept]

    # build the figure
    row_height = 0.4
    plt.figure()
    if plot_size is None:
        plt.gcf().set_size_inches(8, nb_features_kept * row_height + 1.5)
    elif isinstance(plot_size,(list, tuple)):
        plt.gcf().set_size_inches(plot_size[0], plot_size[1])
    elif isinstance(plot_size, float):
        row_height = plot_size
        plt.gcf().set_size_inches(8, nb_features_kept * row_height + 1.5)
    else:
        raise ValueError("Wrong plot_size argument, type can be [None, float, list, tuple]")

    plt.axvline(0, 0, 1, color="dimgray", linestyle="-", linewidth=1)

    # make the beeswarm dots
    for pos, i in enumerate(features_idx_kept):

        plt.axhline(y=pos, color="darkgrey", lw=0.5, dashes=(1, 5), zorder=-1)
        explanation_val_feature = explanations[:, i]

        if features_values is None:
            fvalues = None
        else:
            fvalues = features_values[:, i]

        # we need to set an offstet y position for points having close
        # explanation value in order to avoid points overlapping
        nb_points = len(explanation_val_feature)
        offset = _get_offset_positions(explanation_val_feature, nb_points, row_height)

        # if features values are not provided then make the points all grey
        if fvalues is None:
            plt.scatter(
                explanation_val_feature,
                pos + offset,
                alpha=alpha,
                linewidth=0,
                zorder=3,
                color="slategrey",
                s=16,
                rasterized=True
            )

        else:
            # if needed, apply clip_percentile
            fvalues = _clip_values(fvalues, clip_percentile)

            plt.scatter(
                explanation_val_feature,
                pos + offset,
                alpha=alpha,
                linewidth=0,
                zorder=3,
                cmap=cmap,
                s=16,
                c=fvalues,
                rasterized=True
            )

    # draw the color bar
    if features_values is not None:
        _add_colorbar(cmap)

    # make the plot prettier
    plt.gca().xaxis.set_ticks_position('bottom')
    plt.gca().yaxis.set_ticks_position('none')
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    plt.yticks(range(nb_features_kept), yticklabels, fontsize=13)
    plt.gca().tick_params('y', length=20, width=0.5, which='major')
    plt.gca().tick_params('x', labelsize=11)
    plt.ylim(-1, nb_features_kept)
    plt.xlabel("Impact on output", fontsize=13)
    plt.tight_layout()



import numpy as np

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, GradCAM, Occlusion, Rise, GuidedBackprop, DeconvNet,
                                  GradCAMPP)
from xplique.metrics import MuFidelity, Deletion, Insertion, AverageStability
from xplique.metrics.base import ExplanationMetric, ExplainerMetric
from ..utils import generate_data, generate_model


def _default_methods(model, output_layer_index=-2):
    return [
        Saliency(model, output_layer_index),
        GradientInput(model, output_layer_index),
        SmoothGrad(model, output_layer_index, nb_samples=5),
        VarGrad(model, output_layer_index),
        SquareGrad(model, output_layer_index),
        IntegratedGradients(model, output_layer_index, steps=5),
        GradCAM(model, output_layer_index),
        Occlusion(model, patch_size=4, patch_stride=4),
        Rise(model, nb_samples=10),
        GuidedBackprop(model, output_layer_index),
        DeconvNet(model, output_layer_index),
        GradCAMPP(model, output_layer_index),
    ]


def test_common():
    """Test that all the attributions method works as explainer"""

    input_shape, nb_labels, samples = ((16, 16, 3), 10, 20)
    x, y = generate_data(input_shape, nb_labels, samples)
    model = generate_model(input_shape, nb_labels)

    explainers = _default_methods(model)

    metrics = [
        Deletion(model, x, y, steps=3),
        Insertion(model, x, y, steps=3),
        MuFidelity(model, x, y, nb_samples=3),
        AverageStability(model, x, y, nb_samples=3)
    ]

    for explainer in explainers:
        explanations = explainer(x, y)
        for metric in metrics:
            assert hasattr(metric, 'evaluate')
            if isinstance(metric, ExplainerMetric):
                score = metric(explainer)
            else:
                assert hasattr(metric, 'inference_function')
                assert hasattr(metric, 'batch_inference_function')
                score = metric(explanations)
            assert type(score) in [np.float32, np.float64, float]

def test_add_activation():
    """Test that adding a softmax or sigmoid layer still works"""
    input_shape, nb_labels, samples = ((16, 16, 3), 5, 8)
    x, y = generate_data(input_shape, nb_labels, samples)
    model = generate_model(input_shape, nb_labels)

    explainers = _default_methods(model)

    activations = ["sigmoid", "softmax"]

    for explainer in explainers:
        explanations = explainer(x, y)
        for activation in activations:
            metrics = [
                Deletion(model, x, y, steps=3, activation=activation),
                Insertion(model, x, y, steps=3, activation=activation),
                MuFidelity(model, x, y, nb_samples=3, activation=activation),
            ]
            for metric in metrics:
                assert hasattr(metric, 'evaluate')
                if isinstance(metric, ExplainerMetric):
                    score = metric(explainer)
                else:
                    assert hasattr(metric, 'inference_function')
                    assert hasattr(metric, 'batch_inference_function')
                    score = metric(explanations)
                    if not isinstance(metric, MuFidelity):
                        assert np.all(score <= 1)
                        assert np.all(score >= 0)
                assert type(score) in [np.float32, np.float64, float]
